# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import io
from datetime import datetime, timezone
from pathlib import Path

import pytest
import scipp as sc

from scippneutron.io import cif


def write_to_str(block: cif.Block) -> str:
    buffer = io.StringIO()
    block.write(buffer)
    buffer.seek(0)
    return buffer.getvalue()


def test_write_block_empty():
    block = cif.Block('a-block-name')
    res = write_to_str(block)
    assert res == 'data_a-block-name\n\n'


def test_write_block_name_with_space():
    with pytest.raises(ValueError):
        cif.Block('a block-name with space')


def test_write_block_comment():
    block = cif.Block('a-block-name', comment='some comment\n to describe the block')
    res = write_to_str(block)
    assert (
        res
        == '''# some comment
#  to describe the block
data_a-block-name

'''
    )


def test_write_block_single_pair_string():
    block = cif.Block('single', [{'audit.creation_method': 'written_by_scippneutron'}])
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method written_by_scippneutron
'''
    )


def test_write_block_single_pair_string_variable():
    block = cif.Block(
        'single', [{'audit.creation_method': sc.scalar('written_by_scippneutron')}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method written_by_scippneutron
'''
    )


def test_write_block_single_pair_number():
    block = cif.Block('number', [{'cell.angle_alpha': 62}])
    res = write_to_str(block)
    assert (
        res
        == '''data_number

_cell.angle_alpha 62
'''
    )


@pytest.mark.parametrize('unit', (None, 'deg'))
def test_write_block_single_pair_number_variable(unit):
    block = cif.Block('number', [{'cell.angle_alpha': sc.scalar(93, unit=unit)}])
    res = write_to_str(block)
    assert (
        res
        == '''data_number

_cell.angle_alpha 93
'''
    )


@pytest.mark.parametrize('unit', (None, 'deg'))
def test_write_block_single_pair_number_error(unit):
    block = cif.Block(
        'number', [{'cell.angle_alpha': sc.scalar(93.2, variance=2.1**2, unit=unit)}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_number

_cell.angle_alpha 93(2)
'''
    )


def test_write_block_single_pair_datetime():
    dt = datetime(
        year=2023, month=12, day=1, hour=15, minute=9, second=45, tzinfo=timezone.utc
    )
    block = cif.Block(
        'datetime',
        [
            {
                'audit.creation_date': dt,
            }
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_datetime

_audit.creation_date 2023-12-01T15:09:45+00:00
'''
    )


def test_write_block_single_pair_datetime_variable():
    block = cif.Block(
        'datetime',
        [
            {
                'audit.creation_date': sc.datetime('2023-12-01T15:12:33'),
            }
        ],
    )
    res = write_to_str(block)
    # No timezone info in the output!
    assert (
        res
        == '''data_datetime

_audit.creation_date 2023-12-01T15:12:33
'''
    )


def test_write_block_single_pair_space():
    block = cif.Block('single', [{'audit.creation_method': 'written by scippneutron'}])
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method 'written by scippneutron'
'''
    )


def test_write_block_single_pair_single_quote():
    block = cif.Block(
        'single', [{'audit.creation_method': "written by 'scippneutron'"}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method "written by 'scippneutron'"
'''
    )


def test_write_block_single_pair_double_quote():
    block = cif.Block(
        'single', [{'audit.creation_method': 'written by "scippneutron"'}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method 'written by "scippneutron"'
'''
    )


def test_write_block_single_pair_both_quotes():
    block = cif.Block(
        'single', [{'audit.creation_method': """'written by "scippneutron"'"""}]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method
; 'written by "scippneutron"'
;
'''
    )


def test_write_block_single_pair_newline():
    block = cif.Block(
        'single',
        [{'audit.creation_method': "written by scippneutron\n    version 2000"}],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_single

_audit.creation_method
; written by scippneutron
    version 2000
;
'''
    )


def test_write_block_single_pair_utf8():
    block = cif.Block('utf-8', [{'audit.creation_method': 'Unicode: \xb5\xc5'}])
    res = write_to_str(block)
    assert (
        res
        == r'''data_utf-8

_audit.creation_method 'Unicode: \xb5\xc5'
'''
    )


def test_write_block_single_pair_single_line_comment():
    block = cif.Block('comment')
    block.add({'diffrn_radiation.probe': 'neutron'}, comment='a comment')
    res = write_to_str(block)
    assert (
        res
        == '''data_comment

# a comment
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_single_pair_multi_line_comment():
    block = cif.Block('comment')
    block.add(
        {'diffrn_radiation.probe': 'neutron'},
        comment='Guessing that\nthis is the\n  correct probe',
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_comment

# Guessing that
# this is the
#   correct probe
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_single_pair_single_line_comment_utf8():
    block = cif.Block('comment')
    block.add({'diffrn_radiation.probe': 'neutron'}, comment='unicode: \xc5')
    res = write_to_str(block)
    assert (
        res
        == r'''data_comment

# unicode: \xc5
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_multiple_pairs():
    block = cif.Block(
        'multiple',
        [
            {
                'audit.creation_method': 'written_by_scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            }
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_multiple

_audit.creation_method written_by_scippneutron
_audit.creation_date 2023-12-01T13:52:00Z
'''
    )


def test_write_block_multiple_chunks():
    block = cif.Block(
        'multiple',
        [
            {
                'audit.creation_method': 'written_by_scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            }
        ],
    )
    block.add({'diffrn_radiation.probe': 'neutron'})
    res = write_to_str(block)
    assert (
        res
        == '''data_multiple

_audit.creation_method written_by_scippneutron
_audit.creation_date 2023-12-01T13:52:00Z

_diffrn_radiation.probe neutron
'''
    )


def test_write_block_multiple_chunks_comment():
    block = cif.Block(
        'multiple',
        [
            {
                'audit.creation_method': 'written_by_scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            }
        ],
    )
    block.add({'diffrn_radiation.probe': 'neutron'}, comment='Guessed')
    res = write_to_str(block)
    assert (
        res
        == '''data_multiple

_audit.creation_method written_by_scippneutron
_audit.creation_date 2023-12-01T13:52:00Z

# Guessed
_diffrn_radiation.probe neutron
'''
    )


def test_write_block_single_loop_one_column():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block('looped', [cif.Loop({'diffrn.ambient_environment': env})])
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


def test_write_block_single_loop_one_column_comment():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            cif.Loop(
                {'diffrn.ambient_environment': env},
                comment='This data is completely made up!',
            )
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

# This data is completely made up!
loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


def test_write_block_single_loop_two_columns():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    id_ = sc.array(dims=['x'], values=['123', 'x6a'])
    block = cif.Block(
        'looped', [cif.Loop({'diffrn.ambient_environment': env, 'diffrn.id': id_})]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
_diffrn.id
water 123
sulfur x6a
'''
    )


def test_write_block_single_loop_multi_line_string():
    env = sc.array(dims=['x'], values=['water\nand some salt', 'sulfur'])
    id_ = sc.array(dims=['x'], values=['123', 'x6a'])
    block = cif.Block(
        'looped', [cif.Loop({'diffrn.ambient_environment': env, 'diffrn.id': id_})]
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
_diffrn.id
; water
and some salt
;
123
sulfur
x6a
'''
    )


def test_write_block_single_loop_numbers():
    coeff = sc.array(dims=['cal'], values=[3.65, -0.012, 1.2e-5])
    power = sc.array(dims=['cal'], values=[0, 1, 2])
    id_ = sc.array(dims=['cal'], values=['tzero', 'DIFC', 'DIFA'])
    block = cif.Block(
        'looped',
        [
            cif.Loop(
                {
                    'pd_calib_d_to_tof.id': id_,
                    'pd_calib_d_to_tof.power': power,
                    'pd_calib_d_to_tof.coeff': coeff,
                }
            )
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_pd_calib_d_to_tof.id
_pd_calib_d_to_tof.power
_pd_calib_d_to_tof.coeff
tzero 0 3.65
DIFC 1 -0.012
DIFA 2 1.2e-05
'''
    )


def test_write_block_single_loop_numbers_errors():
    coeff = sc.array(
        dims=['cal'],
        values=[3.65, -0.012, 1.2e-5],
        variances=[0.13**2, 0.001**2, 2e-6**2],
    )
    power = sc.array(dims=['cal'], values=[0, 1, 2])
    id_ = sc.array(dims=['cal'], values=['tzero', 'DIFC', 'DIFA'])
    block = cif.Block(
        'looped',
        [
            cif.Loop(
                {
                    'pd_calib_d_to_tof.id': id_,
                    'pd_calib_d_to_tof.power': power,
                    'pd_calib_d_to_tof.coeff': coeff,
                }
            )
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_pd_calib_d_to_tof.id
_pd_calib_d_to_tof.power
_pd_calib_d_to_tof.coeff
tzero 0 3.65(13)
DIFC 1 -0.0120(10)
DIFA 2 0.000012(2)
'''
    )


def test_write_block_pairs_then_loop():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            {
                'audit.creation_method': 'written by scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            },
            cif.Loop({'diffrn.ambient_environment': env}),
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

_audit.creation_method 'written by scippneutron'
_audit.creation_date 2023-12-01T13:52:00Z

loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


def test_write_block_loop_then_pairs():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            cif.Loop({'diffrn.ambient_environment': env}),
            {
                'audit.creation_method': 'written by scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            },
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
water
sulfur

_audit.creation_method 'written by scippneutron'
_audit.creation_date 2023-12-01T13:52:00Z
'''
    )


def test_write_block_pair_then_loop_then_pairs():
    env = sc.array(dims=['x'], values=['water', 'sulfur'])
    block = cif.Block(
        'looped',
        [
            {'diffrn_radiation.probe': 'neutron'},
            cif.Loop({'diffrn.ambient_environment': env}),
            {
                'audit.creation_method': 'written by scippneutron',
                'audit.creation_date': '2023-12-01T13:52:00Z',
            },
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

_diffrn_radiation.probe neutron

loop_
_diffrn.ambient_environment
water
sulfur

_audit.creation_method 'written by scippneutron'
_audit.creation_date 2023-12-01T13:52:00Z
'''
    )


def test_write_block_two_loops():
    env = sc.array(dims=['env'], values=['water', 'sulfur'])
    author = sc.array(dims=['author'], values=['Ridcully, M.', 'Librarian'])
    email = sc.array(dims=['author'], values=['m.ridcully@uu.am', 'lib@uu.am'])
    block = cif.Block(
        'looped',
        [
            cif.Loop({'diffrn.ambient_environment': env}),
            cif.Loop({'audit_author.name': author, 'audit_author.email': email}),
        ],
    )
    res = write_to_str(block)
    assert (
        res
        == '''data_looped

loop_
_diffrn.ambient_environment
water
sulfur

loop_
_audit_author.name
_audit_author.email
'Ridcully, M.' m.ridcully@uu.am
Librarian lib@uu.am
'''
    )


def test_write_block_core_schema_from_chunk():
    chunk = cif.Chunk(
        {'audit.creation_method': 'written by scippneutron'}, schema=cif.CORE_SCHEMA
    )
    block = cif.Block('block-with-schema', [chunk])
    res = write_to_str(block)
    assert (
        res
        == '''data_block-with-schema

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/fc3d75a298fd7c0c3cde43633f2a8616e826bfd5/cif_core.dic

_audit.creation_method 'written by scippneutron'
'''  # noqa: E501
    )


def test_write_block_core_schema_from_loop():
    author = sc.array(dims=['author'], values=['Ridcully, M.', 'Librarian'])
    email = sc.array(dims=['author'], values=['m.ridcully@uu.am', 'lib@uu.am'])
    loop = cif.Loop(
        {'audit_author.name': author, 'audit_author.email': email},
        schema=cif.CORE_SCHEMA,
    )
    block = cif.Block('block-with-schema', [loop])
    res = write_to_str(block)
    assert (
        res
        == '''data_block-with-schema

loop_
_audit_conform.dict_name
_audit_conform.dict_version
_audit_conform.dict_location
coreCIF 3.3.0 https://github.com/COMCIFS/cif_core/blob/fc3d75a298fd7c0c3cde43633f2a8616e826bfd5/cif_core.dic

loop_
_audit_author.name
_audit_author.email
'Ridcully, M.' m.ridcully@uu.am
Librarian lib@uu.am
'''  # noqa: E501
    )


def test_write_block_pd_schema_from_chunk():
    chunk = cif.Chunk(
        {'pd_meas.units_of_intensity': '1/(micro ampere)'}, schema=cif.PD_SCHEMA
    )
    block = cif.Block('block-with-schema', [chunk])
    res = write_to_str(block)
    # The order of schemas is arbitrary, so we cannot easily check the whole string.
    assert 'pdCIF' in res
    assert 'coreCIF' in res


def test_write_block_multi_schema_schema():
    core_chunk = cif.Chunk(
        {'audit.creation_method': 'written by scippneutron'}, schema=cif.CORE_SCHEMA
    )
    pd_chunk = cif.Chunk(
        {'pd_meas.units_of_intensity': '1/(micro ampere)'}, schema=cif.PD_SCHEMA
    )
    block = cif.Block('block-with-schema', [core_chunk, pd_chunk])
    res = write_to_str(block)
    # The order of schemas is arbitrary, so we cannot easily check the whole string.
    assert 'pdCIF' in res
    assert 'coreCIF' in res


def test_save_cif_one_block_buffer():
    block1 = cif.Block(
        'block-1', [{'audit.creation_method': 'written by scippneutron'}]
    )
    buffer = io.StringIO()
    cif.save_cif(buffer, block1)
    buffer.seek(0)
    assert (
        buffer.read()
        == r'''#\#CIF_1.1
data_block-1

_audit.creation_method 'written by scippneutron'
'''
    )


def test_save_cif_two_blocks_buffer():
    env = sc.array(dims=['env'], values=['water', 'sulfur'])
    block1 = cif.Block(
        'block-1', [{'audit.creation_method': 'written by scippneutron'}]
    )
    block2 = cif.Block(
        'block-2',
        [
            {'diffrn_radiation.probe': 'neutron'},
            cif.Loop({'diffrn.ambient_environment': env}),
        ],
    )
    buffer = io.StringIO()
    cif.save_cif(buffer, [block1, block2])
    buffer.seek(0)
    assert (
        buffer.read()
        == r'''#\#CIF_1.1
data_block-1

_audit.creation_method 'written by scippneutron'

data_block-2

_diffrn_radiation.probe neutron

loop_
_diffrn.ambient_environment
water
sulfur
'''
    )


@pytest.mark.parametrize('path_type', (str, Path))
def test_save_cif_one_block_file(tmpdir, path_type):
    path = path_type(Path(tmpdir) / "test_save_cif_one_block.cif")
    block1 = cif.Block(
        'block-1', [{'audit.creation_method': 'written by scippneutron'}]
    )

    cif.save_cif(path, block1)
    with open(path, 'r') as f:
        assert (
            f.read()
            == r'''#\#CIF_1.1
data_block-1

_audit.creation_method 'written by scippneutron'
'''
        )


def test_loop_requires_1d():
    with pytest.raises(sc.DimensionError):
        cif.Loop({'fake': sc.zeros(sizes={'x': 4, 'y': 3})})


def test_loop_requires_matching_dims():
    with pytest.raises(sc.DimensionError):
        cif.Loop({'a': sc.zeros(sizes={'x': 4}), 'b': sc.zeros(sizes={'x': 3})})
    with pytest.raises(sc.DimensionError):
        cif.Loop({'a': sc.zeros(sizes={'x': 4}), 'b': sc.zeros(sizes={'y': 4})})


def test_block_with_reduced_powder_data():
    da = sc.DataArray(
        sc.array(
            dims=['tof'],
            values=[13.6, 26.0, 9.7],
            variances=[0.81, 1, 0.36],
        ),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')},
    )

    block = cif.Block('reduced', [])
    block.add_reduced_powder_data(da)
    res = write_to_str(block)

    assert 'pdCIF' in res
    assert 'coreCIF' in res

    _, _, tof_loop = res.split('\n\n')
    assert (
        tof_loop
        == '''loop_
_pd_meas.time_of_flight
_pd_proc.intensity_net
_pd_proc.intensity_net_su
1.2 13.6 0.9
1.4 26.0 1.0
2.3 9.7 0.6
'''
    )


def test_block_with_reduced_powder_data_custom_unit():
    da = sc.DataArray(
        sc.array(dims=['tof'], values=[13.6, 26.0, 9.7], unit='counts'),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')},
    )

    block = cif.Block('reduced', [])
    block.add_reduced_powder_data(da)
    res = write_to_str(block)

    assert 'pdCIF' in res
    assert 'coreCIF' in res

    _, _, tof_loop = res.split('\n\n')
    assert (
        tof_loop
        == '''# Unit of intensity: [counts]
loop_
_pd_meas.time_of_flight
_pd_proc.intensity_net
1.2 13.6
1.4 26.0
2.3 9.7
'''
    )


def test_block_with_reduced_powder_data_bad_dim():
    da = sc.DataArray(
        sc.array(
            dims=['time'],
            values=[13.6, 26.0, 9.7],
        ),
        coords={'time': sc.array(dims=['time'], values=[1.2, 1.4, 2.3], unit='us')},
    )

    block = cif.Block('reduced', [])
    with pytest.raises(sc.CoordError):
        block.add_reduced_powder_data(da)


def test_block_with_reduced_powder_data_bad_name():
    da = sc.DataArray(
        sc.array(
            dims=['tof'],
            values=[13.6, 26.0, 9.7],
        ),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='us')},
        name='bad',
    )

    block = cif.Block('reduced', [])
    with pytest.raises(ValueError):
        block.add_reduced_powder_data(da)


def test_block_with_reduced_powder_data_bad_coord_unit():
    da = sc.DataArray(
        sc.array(dims=['tof'], values=[13.6, 26.0, 9.7]),
        coords={'tof': sc.array(dims=['tof'], values=[1.2, 1.4, 2.3], unit='ns')},
    )

    block = cif.Block('reduced', [])
    with pytest.raises(sc.UnitError):
        block.add_reduced_powder_data(da)


def test_block_powder_calibration():
    da = sc.DataArray(
        sc.array(dims=['cal'], values=[1.2, 4.5, 6.7]),
        coords={'power': sc.array(dims=['cal'], values=[0, 1, -1])},
    )
    block = cif.Block('cal', [])
    block.add_powder_calibration(da)
    res = write_to_str(block)

    assert 'pdCIF' in res
    assert 'coreCIF' in res

    _, _, cal_loop = res.split('\n\n')
    assert (
        cal_loop
        == '''loop_
_pd_calib_d_to_tof.id
_pd_calib_d_to_tof.power
_pd_calib_d_to_tof.coeff
tzero 0 1.2
DIFC 1 4.5
DIFB -1 6.7
'''
    )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""
This module uses the confluent-kafka-python implementation of the
Kafka Client API to communicate with Kafka servers (brokers)

The Kafka documentation gives a useful brief introduction you may
want to read:
https://kafka.apache.org/documentation/#gettingStarted

The following Kafka terminology is used extensively in the code:
- A "topic" is a named data stream, messages are published to, or
  consumed from, a topic.
- A "partition" is a logfile of messages on the broker, each topic
  comprises one or more partitions.
- The "offset" is the index of the message in a partition on the
  Kafka broker. Old messages are deleted from the partition, so the
  first available message may not be at offset 0.
"""

import multiprocessing as mp
import threading
from queue import Empty as QueueEmpty
from time import time_ns
from typing import Callable, List, Optional, Tuple, Union
from warnings import warn

import numpy as np
from confluent_kafka import Consumer, KafkaError, TopicPartition
from streaming_data_types.exceptions import WrongSchemaException
from streaming_data_types.run_start_pl72 import RunStartInfo, deserialise_pl72

from ._consumer_type import ConsumerType


class RunStartError(Exception):
    pass


class FakeConsumer:
    """
    Use in place of confluent_kafka.Consumer
    to avoid network io in unit tests
    """

    def __init__(self, input_queue: Optional[mp.Queue]):
        if input_queue is None:
            raise RuntimeError(
                "A multiprocessing queue for test messages "
                "must be provided when using FakeConsumer"
            )
        # This queue is used to provide the consumer with
        # messages in unit tests, instead of it getting messages
        # from the Kafka broker
        self._input_queue = input_queue

    def assign(self, topic_partitions: List[TopicPartition]):
        pass

    def poll(self, timeout: float):
        try:
            msg = self._input_queue.get(timeout=0.1)
            return msg
        except QueueEmpty:
            pass

    def close(self):
        pass


class KafkaConsumer:
    def __init__(
        self,
        topic_partition: TopicPartition,
        consumer: Union[Consumer, FakeConsumer],
        callback: Callable,
        stop_time_ms: Optional[int] = None,
    ):
        self._consumer = consumer
        # To consume messages the consumer must "subscribe" to one
        # or more topics or "assign" specific topic partitions, the
        # latter allows us to start consuming at an offset specified
        # in the TopicPartition. Each KafkaConsumer consumes from only
        # a single partition, this simplifies the logic around run stop
        # behaviour.
        # Note that offsets are integer indices pointing to a position in the
        # topic partition, they are not a bytes offset.
        self._consumer.assign([topic_partition])
        self._callback = callback
        self._reached_eop = False
        self.cancelled = False
        self._consume_data: Optional[threading.Thread] = None
        self.stopped = True
        self._stop_time_mutex = threading.Lock()
        # default stop time to distant future (run until manually stopped)
        self._stop_time = np.iinfo(np.int64).max
        if stop_time_ms is not None:
            self._stop_time = stop_time_ms

    def start(self):
        self.stopped = False
        self.cancelled = False
        self._consume_data = threading.Thread(target=self._consume_loop)
        self._consume_data.start()

    def _consume_loop(self):
        def time_now_ms() -> int:
            return time_ns() // 1_000_000

        reached_message_after_stop_time = False
        reached_stop_time = False
        at_end_of_partition = False

        while not self.cancelled:
            with self._stop_time_mutex:
                if time_now_ms() > self._stop_time:
                    reached_stop_time = True
                    # Wall clock time is after the run stop time.
                    # Now we just continue until we either consume a
                    # message with a timestamp that is after the stop time
                    # or we reach the end of messages available on Kafka
                    # (end of partition error).
                    if reached_message_after_stop_time:
                        # We've already seen a message timestamped
                        # after the stop time.
                        self.cancelled = True
                        break
            msg = self._consumer.poll(timeout=2.0)
            if msg is None:
                if reached_stop_time and at_end_of_partition:
                    self.cancelled = True
                    break
                continue
            if msg.error():
                if msg.error().code() == KafkaError._PARTITION_EOF:
                    at_end_of_partition = True
                    if reached_stop_time:
                        # Wall clock time is after run stop time and there
                        # are no more messages available on Kafka for us to
                        # consume, so cancel running the consumer.
                        self.cancelled = True
                        break
                    continue
                warn(f"Message error in consumer: {msg.error()}")
                self.cancelled = True
                break

            at_end_of_partition = False
            with self._stop_time_mutex:
                if msg.timestamp()[1] > self._stop_time:
                    reached_message_after_stop_time = True
                    if reached_stop_time:
                        # Wall clock time is after run stop time and remaining
                        # messages on Kafka have timestamps after the stop
                        # time, so cancel running the consumer.
                        self.cancelled = True
                        break
                    continue
            self._callback(msg.value())

    def stop(self):
        self.cancelled = True
        if self._consume_data is not None and self._consume_data.is_alive():
            self._consume_data.join(5.0)
        self._consumer.close()
        self.stopped = True

    def update_stop_time(self, new_stop_time_ms: int):
        with self._stop_time_mutex:
            self._stop_time = new_stop_time_ms


class KafkaQueryConsumer:
    """
    Wraps Kafka library consumer methods which query the
    broker for metadata and poll for single messages.
    It is a thin wrapper but allows a fake to be used
    in unit tests.
    """

    def __init__(self, broker: str):
        # Set "enable.auto.commit" to False, as we do not need to report to the
        # kafka broker where we got to (it usually does this in case of a
        # crash, but we simply restart the process and go and find the last
        # run_start message.
        #
        # Set "queued.min.messages" to 1 as we will consume backwards through
        # the partition one message at a time; we do not want to retrieve
        # multiple messages in the forward direction each time we step
        # backwards by 1 offset
        conf = {
            "bootstrap.servers": broker,
            "group.id": "consumer_group_name",
            "auto.offset.reset": "latest",
            "enable.auto.commit": False,
            "queued.min.messages": 1,
        }
        self._consumer = Consumer(**conf)

    def get_topic_partitions(self, topic: str, offset: int = -1):
        metadata = self._consumer.list_topics(topic)
        return [
            TopicPartition(topic, partition[1].id, offset=offset)
            for partition in metadata.topics[topic].partitions.items()
        ]

    def seek(self, partition: TopicPartition):
        """
        Set offset in partition, the consumer will seek to that offset
        """
        self._consumer.seek(partition)

    def poll(self, timeout=2.0):
        """
        Poll for a message from Kafka
        """
        return self._consumer.poll(timeout=timeout)

    def get_watermark_offsets(self, partition: TopicPartition) -> Tuple[int, int]:
        """
        Get the offset of the first and last available
        message in the given partition
        """
        return self._consumer.get_watermark_offsets(partition, cached=False)

    def assign(self, partitions: List[TopicPartition]):
        self._consumer.assign(partitions)

    def offsets_for_times(self, partitions: List[TopicPartition]):
        return self._consumer.offsets_for_times(partitions)


def create_consumers(
    start_time_ms: int,
    stop_time_ms: Optional[int],
    topics: List[str],
    kafka_broker: str,
    consumer_type_enum: ConsumerType,  # so we can inject fake consumer
    callback: Callable,
    test_message_queue: Optional[mp.Queue],
) -> List[KafkaConsumer]:
    """
    Creates one consumer per TopicPartition that start consuming
    at specified timestamp in the data stream

    Having each consumer only be responsible for one partition
    greatly simplifies the logic around stopping at the end of
    the stream (making use of "end of partition" event)
    """
    topic_partitions = []
    if consumer_type_enum == ConsumerType.REAL:
        query_consumer = KafkaQueryConsumer(kafka_broker)
        for topic in topics:
            topic_partitions.extend(
                query_consumer.get_topic_partitions(topic, offset=start_time_ms)
            )
        topic_partitions = query_consumer.offsets_for_times(topic_partitions)

    # Run start messages are typically much larger than the
    # default maximum message size of 1MB. There are
    # corresponding settings on the broker.
    # Note: the "message.max.bytes" does not necessarily have to agree with the
    # size set in the broker. The lower of the two will set the limit.
    # If a message exceeds this maximum size, an error should be reported by
    # the software publishing the run start message (for example NICOS).
    config = {
        "bootstrap.servers": kafka_broker,
        "group.id": "consumer_group_name",
        "auto.offset.reset": "latest",
        "enable.auto.commit": False,
        "message.max.bytes": 100_000_000,
        "fetch.message.max.bytes": 100_000_000,
        "enable.partition.eof": True,  # used by consumer stop logic
    }

    if consumer_type_enum == ConsumerType.REAL:
        consumers = [
            KafkaConsumer(topic_partition, Consumer(config), callback, stop_time_ms)
            for topic_partition in topic_partitions
        ]
    else:
        consumers = [
            KafkaConsumer(
                TopicPartition(""),
                FakeConsumer(test_message_queue),
                callback,
                stop_time_ms,
            )
        ]

    return consumers


def start_consumers(consumers: List[KafkaConsumer]):
    for consumer in consumers:
        consumer.start()


def stop_consumers(consumers: List[KafkaConsumer]):
    for consumer in consumers:
        consumer.stop()


def all_consumers_stopped(consumers: List[KafkaConsumer]) -> bool:
    for consumer in consumers:
        if not consumer.stopped:
            # if the consumer is cancelled then cleanly stop it
            # (join the thread)
            if consumer.cancelled:
                consumer.stop()
                continue
            return False
    return True


def get_run_start_message(
    topic: str, query_consumer: KafkaQueryConsumer
) -> RunStartInfo:
    """
    Get the last run start message on the given topic.

    TODO: we may need to carry out some filtering on instrument name,
    since it is possible that start messages from other instruments end up on
    the same Kafka topic (unless there is a separate topic for each instrument)
    """
    topic_partitions = query_consumer.get_topic_partitions(topic)
    n_partitions = len(topic_partitions)
    if n_partitions != 1:
        raise RuntimeError(
            f"Expected run start topic to contain exactly one partition, "
            f"the specified topic '{topic}' has {n_partitions} partitions."
        )
    partition = topic_partitions[0]
    low_watermark_offset, current_offset = query_consumer.get_watermark_offsets(
        partition
    )
    partition.offset = current_offset
    query_consumer.assign([partition])

    # Consume backwards from the end of the partition
    # until we find a run start message or reach the
    # start of the partition
    while current_offset > low_watermark_offset:
        current_offset -= 1
        partition.offset = current_offset
        query_consumer.seek(partition)
        message = query_consumer.poll(timeout=2.0)
        if message is None:
            raise RunStartError("Timed out when trying to retrieve run start message")
        elif message.error():
            raise RunStartError(
                f"Error encountered consuming run start " f"message: {message.error()}"
            )
        try:
            return deserialise_pl72(message.value())
        except WrongSchemaException:
            # Not a run start message, keep trying
            pass

    raise RunStartError(f"Run start message not found in topic '{topic}'")



import json
import sys
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from typing import Any

import numpy as np
import pytest
import scipp as sc
import scipp.testing
import scippnexus as snx

from scippneutron.io.nexus.load_nexus import json_nexus_group, load_nexus_json_str

sys.path.insert(0, str(Path(__file__).resolve().parent))
from nexus_helpers import NexusBuilder, Source, Stream  # noqa: E402


class JsonNexusExampleLoader:
    BASE_PATH = Path(__file__).resolve().parent / "json_nexus_examples"

    @staticmethod
    @lru_cache
    def _load(path: Path) -> dict[str, Any]:
        with path.open('r') as f:
            return json.load(f)

    def load(self, name: str) -> dict[str, Any]:
        return deepcopy(self._load(self.BASE_PATH.joinpath(name).with_suffix('.json')))

    def __getattr__(self, name: str) -> dict[str, Any]:
        try:
            return self.load(name)
        except FileNotFoundError:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            ) from None


@pytest.fixture(scope="module")
def examples():
    return JsonNexusExampleLoader()


def make_group(json_dict: dict[str, Any]) -> snx.Group:
    return json_nexus_group({'children': [json_dict]})


@pytest.mark.skip("TODO Stream handling with log not implemented")
def test_stream_object_as_transformation_results_in_warning():
    builder = NexusBuilder()
    builder.add_component(Source("source"))
    stream_path = "/entry/streamed_nxlog_transform"
    builder.add_stream(Stream(stream_path))
    builder.add_dataset_at_path("/entry/source/depends_on", stream_path, {})

    with pytest.warns(UserWarning):
        loaded_data, _ = load_nexus_json_str(builder.json_string)

    # A 0 distance translation is used in place of the streamed transformation
    default = [0, 0, 0]
    assert np.allclose(loaded_data["source_position"].values, default)
    assert loaded_data["source_position"].unit == sc.Unit("m")


def test_nexus_json_load_dataset_in_entry(examples):
    group = make_group(examples.entry)
    assert group[()]['entry']['title'] == 'my experiment'
    assert group['entry'][()]['title'] == 'my experiment'
    assert group['entry']['title'][()] == 'my experiment'


def test_nexus_json_load_event_data(examples):
    entry = examples.entry
    entry['children'].append(examples.event_data)
    group = make_group(entry)
    loaded = group[()]['entry']['events_0']

    expected_events = sc.DataArray(
        sc.ones(sizes={'event': 5}, unit='counts', dtype='float32'),
        coords={
            'event_id': sc.array(
                dims=['event'], values=[1, 2, 3, 1, 3], unit=None, dtype='int64'
            ),
            'event_time_offset': sc.array(
                dims=['event'],
                values=[456, 743, 347, 345, 632],
                unit='ns',
                dtype='int64',
            ),
        },
    )
    event_time_zero = sc.datetimes(
        dims=['event_time_zero'],
        values=[
            1600773930000000000,
            1600773931000000000,
            1600773932000000000,
            1600773933000000000,
        ],
        unit='ns',
    )
    begin = sc.array(dims=['event'], values=[0, 3, 3, 5], unit=None, dtype='int64')
    end = sc.array(dims=['event'], values=[3, 3, 5, 5], unit=None, dtype='int64')
    binned = sc.bins(data=expected_events, dim='event', begin=begin, end=end)
    expected = sc.DataArray(
        binned.rename_dims({'event': 'event_time_zero'}),
        coords={'event_time_zero': event_time_zero},
    )

    sc.testing.assert_identical(loaded, expected)


def test_nexus_json_load_detector_with_event_data(examples):
    detector = examples.detector
    detector['children'].append(examples.event_data)
    entry = examples.entry
    entry['children'].append(detector)
    group = make_group(entry)
    loaded = group[()]['entry']['detector_0']

    expected_events = sc.DataArray(
        sc.ones(sizes={'event': 5}, unit='counts', dtype='float32'),
        coords={
            'event_id': sc.array(
                dims=['event'], values=[1, 2, 3, 1, 3], unit=None, dtype='int64'
            ),
            'event_time_zero': sc.datetimes(
                dims=['event'],
                values=[
                    1600773930000000000,
                    1600773930000000000,
                    1600773930000000000,
                    1600773932000000000,
                    1600773932000000000,
                ],
                unit='ns',
            ),
            'event_time_offset': sc.array(
                dims=['event'],
                values=[456, 743, 347, 345, 632],
                unit='ns',
                dtype='int64',
            ),
        },
    )
    detector_number = sc.array(
        dims=['detector_number'], values=[1, 2, 3, 4], unit=None, dtype='int64'
    )
    expected = expected_events.group(
        sc.array(dims=['event_id'], values=[1, 2, 3, 4], unit=None, dtype='int64')
    )
    expected = expected.rename({'event_id': 'detector_number'}).assign_coords(
        detector_number=detector_number
    )

    assert loaded.keys() == {'events_0'}
    sc.testing.assert_identical(loaded['events_0'], expected)


def test_nexus_json_load_log(examples):
    entry = examples.entry
    entry['children'].append(examples.log)
    group = make_group(entry)
    loaded = group[()]['entry']['test_log']

    expected = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3], unit='m'),
        coords={
            'time': sc.datetimes(
                dims=['time'],
                values=[1354291220000000000, 1354291231000000000, 1354291242000000000],
                unit='ns',
            )
        },
    )
    sc.testing.assert_identical(loaded, expected)


def test_nexus_json_load_log_utf8_unit(examples):
    log = examples.log
    assert log['children'][0]['config']['name'] == 'value', 'is the expected child'
    assert log['children'][0]['attributes'][0]['name'] == 'units'
    log['children'][0]['attributes'][0]['values'] = '\u00b0'  # '°', i.e., degrees

    entry = examples.entry
    entry['children'].append(log)
    group = make_group(entry)
    loaded = group[()]['entry']['test_log']

    expected = sc.DataArray(
        sc.array(dims=['time'], values=[1.1, 2.2, 3.3], unit='deg'),
        coords={
            'time': sc.datetimes(
                dims=['time'],
                values=[1354291220000000000, 1354291231000000000, 1354291242000000000],
                unit='ns',
            )
        },
    )
    sc.testing.assert_identical(loaded, expected)


# TODO: remove warning filter
# Loading a NXmonitor fails with
# 'Signal is not array like' error and issues a warning.
# But this should probably be fixed elsewhere.
@pytest.mark.filterwarnings('ignore::UserWarning')
def test_nexus_json_load_ymir_instrument(examples):
    make_group(examples.instrument)[()]


def test_nexus_json_load_dataset(examples):
    dg = make_group(examples.dataset)[()]
    assert dg['name'] == 'YMIR'


def test_nexus_json_load_array_dataset(examples):
    dg = make_group(examples.array_dataset)[()]
    sc.testing.assert_identical(
        dg['slit_edges'],
        sc.array(dims=['dim_0'], values=[0.0, 15.0, 180.0, 195.0], unit='deg'),
    )



from mantid.kernel import ConfigService

import scippneutron as scn

ConfigService.setLogLevel(1)


class TransformCoords:
    def setup(self):
        da = scn.load_with_mantid(scn.data.get_path('PG3_4844_event.nxs'))
        self.var_tof = da
        self.var_wavelength = scn.convert(self.var_tof, "tof", "wavelength", False)
        self.var_energy = scn.convert(self.var_tof, "tof", "energy", False)

    def time_wavelength_to_tof(self):
        scn.convert(self.var_tof, "tof", "wavelength", False)

    def time_tof_to_wavelength(self):
        scn.convert(self.var_wavelength, "wavelength", "tof", False)

    def time_tof_to_dspacing(self):
        scn.convert(self.var_tof, "tof", "dspacing", True)

    def time_tof_to_energy(self):
        scn.convert(self.var_tof, "tof", "energy", False)

    def time_energy_to_tof(self):
        scn.convert(self.var_energy, "energy", "tof", False)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import scipp as sc

from ..mantid import load_with_mantid

_version = '5'

__all__ = ['tutorial_dense_data', 'tutorial_event_data', 'get_path']


def _make_pooch():
    import pooch

    return pooch.create(
        path=pooch.os_cache('scippneutron'),
        env='SCIPPNEUTRON_DATA_DIR',
        retry_if_failed=3,
        base_url='https://public.esss.dk/groups/scipp/scippneutron/{version}/',
        version=_version,
        registry={
            'bigfake.nxs': 'md5:ebf73017ddd03cee2a14a887c439bf35',
            'iris26176_graphite002_sqw.nxs': 'md5:7ea63f9137602b7e9b604fe30f0c6ec2',
            'loki-at-larmor.hdf5': 'md5:9c93f36fa2e94a7b63f2b946705e74f3',
            'loki-at-larmor-filtered.hdf5': 'md5:2789d0384339d1e130b86a44f22d11c6',
            'powder-event.h5': 'md5:38f7285b9814d00b5c5b3e63d4455e55',
            'powder-event-filtered.h5': 'md5:2eece38997aa800ec96e4105f85c11bd',
            'mcstas_sans.h5': 'md5:a608255dd757905490882eb35e209d12',
            'CNCS_51936_event.nxs': 'md5:5ba401e489260a44374b5be12b780911',
            'GEM40979.raw': 'md5:6df0f1c2fc472af200eec43762e9a874',
            'PG3_4844_calibration.h5': 'md5:39005d7b4befcafa173f00645549e2b4',
            'PG3_4844_event.nxs': 'md5:d5ae38871d0a09a28ae01f85d969de1e',
            'PG3_4866_event.nxs': 'md5:3d543bc6a646e622b3f4542bc3435e7e',
            'PG3_4871_event.nxs': 'md5:a3d0edcb36ab8e9e3342cd8a4440b779',
            'WISH00016748.raw': 'md5:37ecc6f99662b57e405ed967bdc068af',
        },
    )


_pooch = _make_pooch()


def bigfake():
    return get_path('bigfake.nxs')


def tutorial_dense_data():
    # This file was produced from loki-at-larmor.hdf5 by converting the data into a
    # data group. But not all attributes were copied.
    return sc.io.load_hdf5(_pooch.fetch('loki-at-larmor-filtered.hdf5'))


def tutorial_event_data():
    # This file was produced from powder-event.h5 by converting the data into a
    # data group. But not all attributes were copied.
    return sc.io.load_hdf5(_pooch.fetch('powder-event-filtered.h5'))


def powder_sample():
    return load_with_mantid(_pooch.fetch('PG3_4844_event.nxs'))


def powder_calibration():
    return sc.io.load_hdf5(_pooch.fetch('PG3_4844_calibration.h5'))


def get_path(name: str) -> str:
    """
    Return the path to a data file bundled with scippneutron.

    This function only works with example data and cannot handle
    paths to custom files.
    """
    return _pooch.fetch(name)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa

import os

from .conversions import conversion_graph, convert, deduce_conversion_graph



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Neil Vaytet and Owen Arnold

import numpy as np
import scipp as sc
from scipy.spatial.transform import Rotation as Rot

from ._utils import get_meta

try:
    import pythreejs as p3
except ImportError as ex:
    p3 = None
    _pythreejs_import_error = ex


def _create_text_sprite(position, bounding_box, display_text):
    # Position offset in y
    text_position = tuple(position.value + np.array([0, 0.8 * bounding_box[1], 0]))
    text = p3.TextTexture(string=display_text, color='black', size=300)
    text_material = p3.SpriteMaterial(map=text, transparent=True)
    size = 1.0
    return p3.Sprite(
        material=text_material, position=text_position, scale=[size, size, size]
    )


def _create_mesh(geometry, color, wireframe, position):
    if wireframe:
        edges = p3.EdgesGeometry(geometry)
        mesh = p3.LineSegments(
            geometry=edges, material=p3.LineBasicMaterial(color=color)
        )

    else:
        material = p3.MeshBasicMaterial(color=color)
        mesh = p3.Mesh(geometry=geometry, material=material)
    mesh.position = tuple(position.value)
    return mesh


def _box(position, display_text, bounding_box, color, wireframe, **kwargs):
    geometry = p3.BoxGeometry(
        width=bounding_box[0],
        height=bounding_box[1],
        depth=bounding_box[2],
        widthSegments=2,
        heightSegments=2,
        depthSegments=2,
    )

    mesh = _create_mesh(
        geometry=geometry, color=color, wireframe=wireframe, position=position
    )
    text_mesh = _create_text_sprite(
        position=position, bounding_box=bounding_box, display_text=display_text
    )
    return mesh, text_mesh


def _find_beam(det_com, pos):
    # Assume beam is axis aligned and follows largest axis
    # delta between component and detector COM
    beam_dir = np.argmax(np.abs(det_com.value - pos.value))
    beam = np.zeros(3)
    beam[beam_dir] = 1
    return beam


def _alignment_matrix(to_align, target):
    rot_axis = np.cross(to_align, target)
    magnitude = np.linalg.norm(to_align) * np.linalg.norm(target)
    axis_angle = np.arcsin(rot_axis / magnitude)
    return Rot.from_rotvec(axis_angle).as_matrix()


def _disk_chopper(position, display_text, bounding_box, color, wireframe, **kwargs):
    geometry = p3.CylinderGeometry(
        radiusTop=bounding_box[0] / 2,
        radiusBottom=bounding_box[0] / 2,
        height=bounding_box[0] / 100,
        radialSegments=24,
        heightSegments=12,
        openEnded=False,
        thetaStart=np.pi / 8,
        thetaLength=2 * np.pi - (np.pi / 8),
    )

    mesh = _create_mesh(
        geometry=geometry, color=color, wireframe=wireframe, position=position
    )
    beam = _find_beam(det_com=kwargs['det_center'], pos=position)
    disk_axis = np.array([0, 1, 0])  # Disk created with this axis
    rotation = _alignment_matrix(to_align=disk_axis, target=beam)
    mesh.setRotationFromMatrix(rotation.flatten())
    text_mesh = _create_text_sprite(
        position=position, bounding_box=bounding_box, display_text=display_text
    )
    return mesh, text_mesh


def _cylinder(position, display_text, bounding_box, color, wireframe, **kwargs):
    geometry = p3.CylinderGeometry(
        radiusTop=bounding_box[0] / 2,
        radiusBottom=bounding_box[0] / 2,
        height=bounding_box[1],
        radialSegments=12,
        heightSegments=12,
        openEnded=False,
        thetaStart=0,
        thetaLength=2.0 * np.pi,
    )
    mesh = _create_mesh(
        geometry=geometry, color=color, wireframe=wireframe, position=position
    )
    # Position label above cylinder
    text_mesh = _create_text_sprite(
        position=position, bounding_box=bounding_box, display_text=display_text
    )
    return mesh, text_mesh


def _unpack_to_scene(scene, items):
    if hasattr(items, "__iter__"):
        for item in items:
            scene.add(item)
    else:
        scene.add(items)


def _add_to_scene(
    position, scene, shape, display_text, bounding_box, color, wireframe, **kwargs
):
    _unpack_to_scene(
        scene,
        shape(
            position,
            display_text=display_text,
            bounding_box=bounding_box,
            color=color,
            wireframe=wireframe,
            **kwargs,
        ),
    )


def _furthest_component(det_center, scipp_obj, additional):
    distances = [
        sc.norm(settings["center"] - det_center).value
        for settings in list(additional.values())
    ]
    max_displacement = sorted(distances)[-1]
    return max_displacement


def _instrument_view_shape_types():
    return {"box": _box, "cylinder": _cylinder, "disk": _disk_chopper}


def _as_vector(var):
    if var.dtype == sc.DType.vector3:
        return var
    else:
        return sc.spatial.as_vectors(x=var, y=var, z=var)


def _plot_components(scipp_obj, components, positions_var, scene):
    det_center = sc.mean(positions_var)
    # Some scaling to set width according to distance from detector center
    shapes = _instrument_view_shape_types()
    for name, settings in components.items():
        type = settings["type"]
        size = _as_vector(settings["size"])
        component_position = settings["center"]
        color = settings.get("color", "#808080")
        wireframe = settings.get("wireframe", False)
        if type not in shapes:
            supported_shapes = ", ".join(shapes.keys())
            raise ValueError(
                f"Unknown shape: {type} requested for {name}. "
                f"Allowed values are: {supported_shapes}"
            )
        component_position = sc.to_unit(component_position, positions_var.unit)
        size = sc.to_unit(size, positions_var.unit)
        _add_to_scene(
            position=component_position,
            scene=scene,
            shape=shapes[type],
            display_text=name,
            bounding_box=tuple(size.values),
            color=color,
            wireframe=wireframe,
            det_center=det_center,
        )
    # Reset camera
    camera = _get_camera(scene)
    if camera:
        furthest_distance = _furthest_component(det_center, scipp_obj, components)
        camera.far = max(camera.far, furthest_distance * 5.0)


def _get_camera(scene):
    for child in scene.children:
        if isinstance(child, p3.PerspectiveCamera):
            return child
    return None


def instrument_view(
    scipp_obj, positions="position", pixel_size=None, components=None, **kwargs
):
    """Plot a 3D view of the instrument, using the `position` coordinate as the
    detector vector positions.

    Use the `positions` argument to change the vectors used as pixel positions.
    Sliders are added to navigate extra dimensions.
    Spatial slicing and pixel opacity control is available using the controls
    below the scene.
    Use the `pixel_size` argument to specify the size of the detectors.
    If no `pixel_size` is given, a guess is performed based on the distance
    between the positions of the first two pixel positions.
    The aspect ratio of the positions is preserved by default, but this can
    be changed to automatic scaling using `aspect="equal"`.

    `components` dictionary uses the key as the name to display the component.
    This can be any desired name, it does not have to relate to the input
    `scipp_obj` naming.
    The value for each entry is itself a dictionary that provides the display
    settings and requires:

    * `center` - scipp scalar vector describing position to place item at.
    * `size` - scipp scalar vector describing the bounding box to use in the
    same length units as positions
    * `type` - known shape type to use.
    Valid types are: 'box', 'cylinder' or 'disk'.

    Optional arguments are:

    * `color` - a hexadecimal color code such as #F00000 to use as fill or
    line color
    * `wireframe` - wireframe is a bool that defaults to False. If set to True,
    the returned geometrical shape is a wireframe instead of a shape with
    opaque faces

    Parameters
    ----------
    scipp_obj:
        Scipp object holding geometries.
    positions:
        Key for coord/attr holding positions to use for pixels.
    pixel_size:
        Custom pixel size to use for detector pixels.
    components:
        Dictionary containing display names and corresponding settings
        (also a Dictionary) for additional components to display.
    kwargs:
        Additional keyword arguments to pass to :func:`plopp.scatter3d`.

    Returns
    -------
    :
        The 3D plot object
    """
    if not p3:
        raise _pythreejs_import_error

    import plopp as pp

    positions_var = get_meta(scipp_obj)[positions]
    if pixel_size is None:
        pos_array = positions_var.values
        if len(pos_array) > 1:
            pixel_size = np.linalg.norm(pos_array[1] - pos_array[0])

    fig = pp.scatter3d(scipp_obj, pos=positions, pixel_size=pixel_size, **kwargs)
    scene = fig.children[0].canvas.scene

    # Add additional components from the beamline
    if components:
        _plot_components(scipp_obj, components, positions_var, scene)

    return fig



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import List

import tomli

parser = ArgumentParser()
parser.add_argument(
    "--nightly",
    default="",
    help="List of dependencies to install from main branch for nightly tests, "
    "separated by commas.",
)
args = parser.parse_args()

CUSTOM_AUTO_SEPARATOR = """
# --- END OF CUSTOM SECTION ---
# The following was generated by 'tox -e deps', DO NOT EDIT MANUALLY!
"""


def write_dependencies(dependency_name: str, dependencies: List[str]) -> None:
    path = Path(f"{dependency_name}.in")
    if path.exists():
        sections = path.read_text().split(CUSTOM_AUTO_SEPARATOR)
        if len(sections) > 1:
            custom = sections[0]
        else:
            custom = ""
    else:
        custom = ""
    with path.open("w") as f:
        f.write(custom)
        f.write(CUSTOM_AUTO_SEPARATOR)
        f.write("\n".join(dependencies))
        f.write("\n")


with open("../pyproject.toml", "rb") as toml_file:
    pyproject = tomli.load(toml_file)
    dependencies = pyproject["project"].get("dependencies")
    if not dependencies:
        raise RuntimeError("No dependencies found in pyproject.toml")
    dependencies = [dep.strip().strip('"') for dep in dependencies]

write_dependencies("base", dependencies)


def as_nightly(repo: str) -> str:
    if "/" in repo:
        org, repo = repo.split("/")
    else:
        org = "scipp"
    if repo == "scipp":
        version = f"cp{sys.version_info.major}{sys.version_info.minor}"
        base = "https://github.com/scipp/scipp/releases/download/nightly/scipp-nightly"
        suffix = "manylinux_2_17_x86_64.manylinux2014_x86_64.whl"
        return "-".join([base, version, version, suffix])
    return f"{repo} @ git+https://github.com/{org}/{repo}@main"


nightly = tuple(args.nightly.split(",") if args.nightly else [])
nightly_dependencies = [dep for dep in dependencies if not dep.startswith(nightly)]
nightly_dependencies += [as_nightly(arg) for arg in nightly]

write_dependencies("nightly", nightly_dependencies)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

import json
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

import h5py
import numpy as np
import scipp as sc

from scippneutron.io.nexus._json_nexus import (
    JSONGroup,
    _name,
    make_json_attr,
    make_json_dataset,
)

h5root = Union[h5py.File, h5py.Group]


def _create_nx_class(group_name: str, nx_class_name: str, parent: h5root) -> h5py.Group:
    nx_class = parent.create_group(group_name)
    nx_class.attrs["NX_class"] = nx_class_name
    return nx_class


@contextmanager
def in_memory_hdf5_file_with_two_nxentry() -> Iterator[h5py.File]:
    nexus_file = h5py.File(
        'in_memory_events.nxs', mode='w', driver="core", backing_store=False
    )
    try:
        _create_nx_class("entry_1", "NXentry", nexus_file)
        _create_nx_class("entry_2", "NXentry", nexus_file)
        yield nexus_file
    finally:
        nexus_file.close()


@dataclass
class EventData:
    event_id: Optional[np.ndarray]
    event_time_offset: Optional[np.ndarray]
    event_time_zero: Optional[np.ndarray]
    event_index: Optional[np.ndarray]
    event_time_zero_unit: Optional[Union[str, bytes]] = "ns"
    event_time_zero_offset: Optional[Union[str, bytes]] = "1970-01-01T00:00:00Z"
    event_time_offset_unit: Optional[Union[str, bytes]] = "ns"


@dataclass
class Log:
    name: str
    value: Optional[np.ndarray]
    time: Optional[np.ndarray] = None
    value_units: Optional[Union[str, bytes]] = None

    # From
    # https://manual.nexusformat.org/classes/base_classes/NXlog.html?highlight=nxlog
    # time units are non-optional if time series data is present, and the unit
    # must be a unit of time (i.e. convertible to seconds).
    time_units: Optional[Union[str, bytes]] = "s"

    start_time: Optional[Union[str, bytes]] = "1970-01-01T00:00:00Z"
    scaling_factor: Optional[float] = None


class TransformationType(Enum):
    TRANSLATION = "translation"
    ROTATION = "rotation"


@dataclass
class Transformation:
    transform_type: TransformationType
    vector: np.ndarray
    value: Optional[np.ndarray]
    time: Optional[np.ndarray] = None
    depends_on: Union["Transformation", str, None] = None
    offset: Optional[np.ndarray] = None
    offset_unit: Optional[str] = None
    value_units: Optional[Union[str, bytes]] = None
    time_units: Optional[Union[str, bytes]] = None


@dataclass
class Detector:
    detector_numbers: Optional[np.ndarray] = None
    event_data: Optional[EventData] = None
    log: Optional[Log] = None
    x_offsets: Optional[np.ndarray] = None
    y_offsets: Optional[np.ndarray] = None
    z_offsets: Optional[np.ndarray] = None
    offsets_unit: Optional[Union[str, bytes]] = None
    depends_on: Optional[Transformation] = None
    data: Optional[np.ndarray] = None


@dataclass
class Data:
    name: str
    data: sc.DataArray
    attrs: dict = None


@dataclass
class Sample:
    name: str
    depends_on: Optional[Transformation] = None
    distance: Optional[float] = None
    distance_units: Optional[Union[str, bytes]] = None
    ub_matrix: Optional[np.ndarray] = None
    orientation_matrix: Optional[np.ndarray] = None


@dataclass
class Source:
    name: str
    depends_on: Union[Transformation, None, str] = None
    distance: Optional[float] = None
    distance_units: Optional[Union[str, bytes]] = None


@dataclass
class Chopper:
    name: str
    distance: float
    rotation_speed: float
    distance_units: Optional[str] = None
    rotation_units: Optional[str] = None


@dataclass
class Link:
    new_path: str
    target_path: str


@dataclass
class DatasetAtPath:
    path: str
    data: np.ndarray
    attributes: Dict[str, Any]


@dataclass
class Stream:
    """
    Only present in the JSON NeXus file templates, not in HDF5 NeXus files.
    Records where to find data in Kafka that are streamed during an experiment.
    """

    # Where the builder should place the stream object
    path: str

    # The following members correspond to fields in stream object.
    # Some of them may not be of interest to Scipp but are to other
    # software which consume the json template, for example
    # the Filewriter (https://github.com/ess-dmsc/kafka-to-nexus)

    # Kafka topic (named data stream)
    topic: str = "motion_devices_topic"
    # Source name, allows filtering and multiplexing to different
    # writer_modules by the filewriter software
    source: str = "linear_axis"
    # Tells filewriter which plugin to use to deserialise
    # messages in this stream and how to write the data to file.
    # For example the "f142" writer module deserialises messages which
    # were serialised with the "f142" flatbuffer schema
    # (https://github.com/ess-dmsc/streaming-data-types/) and
    # writes resulting timeseries data to file as an NXlog
    # (https://manual.nexusformat.org/classes/base_classes/NXlog.html)
    writer_module: str = "f142"
    # Deserialised values are expected to be of this type
    type: str = "double"
    # Values have these units
    value_units: str = "m"


@dataclass
class Monitor:
    name: str
    data: np.ndarray
    axes: List[Tuple[str, np.ndarray]]
    events: Optional[EventData] = None
    depends_on: Optional[Transformation] = None


class InMemoryNeXusWriter:
    def add_dataset_at_path(
        self, file_root: h5py.File, path: str, data: np.ndarray, attributes: Dict
    ):
        path_split = path.split("/")
        dataset_name = path_split[-1]
        parent_path = "/".join(path_split[:-1])
        dataset = self.add_dataset(file_root[parent_path], dataset_name, data)
        for name, value in attributes.items():
            self.add_attribute(dataset, name, value)

    @staticmethod
    def add_dataset(
        parent: h5py.Group, name: str, data: Union[str, bytes, np.ndarray]
    ) -> h5py.Dataset:
        return parent.create_dataset(name, data=data)

    @staticmethod
    def add_attribute(
        parent: Union[h5py.Group, h5py.Dataset],
        name: str,
        value: Union[str, bytes, np.ndarray],
    ):
        parent.attrs[name] = value

    @staticmethod
    def add_group(parent: h5py.Group, name: str) -> h5py.Group:
        return parent.create_group(name)

    @staticmethod
    def add_hard_link(file_root: h5py.File, new_path: str, target_path: str):
        try:
            _ = file_root[new_path]
            del file_root[new_path]
        except KeyError:
            pass
        file_root[new_path] = file_root[target_path]

    @staticmethod
    def add_soft_link(file_root: h5py.File, new_path: str, target_path: str):
        try:
            _ = file_root[new_path]
            del file_root[new_path]
        except KeyError:
            pass
        file_root[new_path] = h5py.SoftLink(target_path)


def _get_child(obj, name):
    children = obj["children"]
    for child in children:
        if _name(child) == name:
            return child
    return None


def _get_object_by_path(file_root, path):
    if not path.startswith('/'):
        return _get_child(file_root, path)
    path = path.split('/')[1:]  # Trim leading slash
    obj = file_root
    for name in path:
        obj = _get_child(obj, name)
    return obj


def _add_link_to_json(file_root: Dict, new_path: str, target_path: str):
    new_path_split = new_path.split("/")
    link_name = new_path_split[-1]
    parent_path = "/".join(new_path_split[:-1])
    parent_group = _get_object_by_path(file_root, parent_path)
    link = {"module": "link", "config": {"name": link_name, "target": target_path}}
    existing_object = _get_object_by_path(parent_group, link_name)
    if existing_object is not None:
        parent_group["children"].remove(existing_object)
    parent_group["children"].append(link)


def _parent_and_name_from_path(file_root: Dict, path: str) -> Tuple[Dict, str]:
    path_split = path.split("/")
    name = path_split[-1]
    parent_path = '/'.join(path_split[:-1])
    parent_group = _get_object_by_path(file_root, parent_path)
    return parent_group, name


class JsonWriter:
    def add_dataset_at_path(
        self, file_root: Dict, path: str, data: np.ndarray, attributes: Dict
    ):
        parent_group, dataset_name = _parent_and_name_from_path(file_root, path)
        dataset = self.add_dataset(parent_group, dataset_name, data)
        for name, value in attributes.items():
            self.add_attribute(dataset, name, value)

    @staticmethod
    def add_dataset(
        parent: Dict, name: str, data: Union[str, bytes, np.ndarray]
    ) -> Dict:
        dataset = make_json_dataset(name, data)
        parent["children"].append(dataset)
        return dataset

    @staticmethod
    def add_attribute(
        parent: Dict, name: str, value: Union[str, bytes, list, np.ndarray]
    ):
        attr = make_json_attr(name, value)
        parent["attributes"].append(attr)

    @staticmethod
    def add_group(parent: Dict, name: str) -> Dict:
        new_group = {"type": "group", "name": name, "children": [], "attributes": []}
        parent["children"].append(new_group)
        return new_group

    @staticmethod
    def add_hard_link(file_root: Dict, new_path: str, target_path: str):
        _add_link_to_json(file_root, new_path, target_path)

    @staticmethod
    def add_soft_link(file_root: Dict, new_path: str, target_path: str):
        _add_link_to_json(file_root, new_path, target_path)

    def add_stream(self, file_root: Dict, stream: Stream):
        new_stream = {
            "module": stream.writer_module,
            "config": {
                "topic": stream.topic,
                "source": stream.source,
                "type": stream.type,
                "value_units": stream.value_units,
            },
        }
        if (group := _get_object_by_path(file_root, stream.path)) is None:
            parent, name = _parent_and_name_from_path(file_root, stream.path)
            group = self.add_group(parent, name)
        group["children"].append(new_stream)


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


class NexusBuilder:
    """
    Allows building an in-memory NeXus file for use in tests
    """

    def __init__(self):
        self._event_data: List[EventData] = []
        self._detectors: List[Detector] = []
        self._logs: List[Log] = []
        self._instrument_name: Optional[str] = None
        self._choppers: List[Chopper] = []
        self._title: Optional[str] = None
        self._start_time: Optional[str] = None
        self._end_time: Optional[str] = None
        self._sample: List[Sample] = []
        self._source: List[Source] = []
        self._hard_links: List[Link] = []
        self._soft_links: List[Link] = []
        self._writer = None
        self._datasets: List[DatasetAtPath] = []
        self._streams = []
        self._monitors = []
        self._datas = []

    def add_dataset_at_path(self, path: str, data: np.ndarray, attributes: Dict):
        self._datasets.append(DatasetAtPath(path, data, attributes))

    def _write_datasets(self, root: Union[Dict, h5py.File]):
        for dataset in self._datasets:
            self._writer.add_dataset_at_path(
                root, dataset.path, dataset.data, dataset.attributes
            )

    def add_stream(self, stream: Stream):
        self._streams.append(stream)

    def add_detector(self, detector: Detector):
        self._detectors.append(detector)

    def add_data(self, data: Data):
        self._datas.append(data)

    def add_event_data(self, event_data: EventData):
        self._event_data.append(event_data)

    def add_log(self, log: Log):
        self._logs.append(log)

    def add_instrument(self, name: str):
        self._instrument_name = name

    def add_chopper(self, chopper: Chopper):
        self._choppers.append(chopper)

    def add_title(self, title: str):
        self._title = title

    def add_run_start_time(self, start_time: str):
        self._start_time = start_time

    def add_run_end_time(self, end_time: str):
        self._end_time = end_time

    def add_sample(self, sample: Sample):
        self._sample.append(sample)

    def add_source(self, source: Source):
        self._source.append(source)

    def add_hard_link(self, link: Link):
        """
        If there is a group or dataset at the link path it will
        be replaced by the link
        """
        self._hard_links.append(link)

    def add_soft_link(self, link: Link):
        """
        If there is a group or dataset at the link path it will
        be replaced by the link
        """
        self._soft_links.append(link)

    def add_component(self, component: Union[Sample, Source]):
        # This is a little ugly, but allows parametrisation
        # of tests which should work for sample and source
        if isinstance(component, Sample):
            self.add_sample(component)
        elif isinstance(component, Source):
            self.add_source(component)

    def add_monitor(self, monitor: Monitor):
        self._monitors.append(monitor)

    @property
    def json_string(self):
        self._writer = JsonWriter()
        root = {"children": []}
        self._write_file(root)
        return json.dumps(root, indent=4, cls=NumpyEncoder)

    def create_json_file(self):
        """
        Create a file on disk, do not use this in tests, it is intended to
        be used as a tool during test development
        """
        self._writer = JsonWriter()
        root = {"children": []}
        self._write_file(root)
        with open("test_json.txt", "w") as json_file:
            return json.dump(root, json_file, indent=4, cls=NumpyEncoder)

    @contextmanager
    def json(self):
        try:
            yield JSONGroup(json.loads(self.json_string))
        finally:
            pass

    @contextmanager
    def file(self) -> Iterator[h5py.File]:
        # "core" driver means file is "in-memory" not on disk.
        # backing_store=False prevents file being written to
        # disk on flush() or close().
        nexus_file = h5py.File(
            'in_memory_events.nxs', mode='w', driver="core", backing_store=False
        )
        self._writer = InMemoryNeXusWriter()
        try:
            self._write_file(nexus_file)
            yield nexus_file
        finally:
            nexus_file.close()

    def _write_file(self, nexus_file: Union[h5py.File, Dict]):
        entry_group = self._create_nx_class("entry", "NXentry", nexus_file)
        if self._title is not None:
            self._writer.add_dataset(entry_group, "title", data=self._title)
        if self._start_time is not None:
            self._writer.add_dataset(entry_group, "start_time", data=self._start_time)
        if self._end_time is not None:
            self._writer.add_dataset(entry_group, "end_time", data=self._end_time)
        self._write_event_data(entry_group)
        self._write_logs(entry_group)
        self._write_sample(entry_group)
        self._write_source(entry_group)
        if self._instrument_name is None:
            parent_group = entry_group
            parent_path = "/entry"
        else:
            parent_group = self._write_instrument(entry_group)
            parent_path = "/entry/instrument"
        self._write_choppers(parent_group)
        self._write_detectors(parent_group, parent_path)
        self._write_datasets(nexus_file)
        self._write_streams(nexus_file)
        self._write_links(nexus_file)
        self._write_monitors(nexus_file)
        self._write_datas(parent_group)

    def create_file_on_disk(self, filename: str):
        """
        Create a file on disk, do not use this in tests, it is intended to
        be used as a tool during test development. Output file can be
        explored using a tool such as HDFView.
        """
        nexus_file = h5py.File(filename, mode='w')
        self._writer = InMemoryNeXusWriter()
        try:
            self._write_file(nexus_file)
        finally:
            nexus_file.close()

    def _write_links(self, file_root: Union[h5py.Group, Dict]):
        for hard_link in self._hard_links:
            self._writer.add_hard_link(
                file_root, hard_link.new_path, hard_link.target_path
            )
        for soft_link in self._soft_links:
            self._writer.add_soft_link(
                file_root, soft_link.new_path, soft_link.target_path
            )

    def _write_sample(self, parent_group: Union[h5py.Group, Dict]):
        for sample in self._sample:
            sample_group = self._create_nx_class(sample.name, "NXsample", parent_group)
            if sample.depends_on is not None:
                depends_on = self._add_transformations_to_file(
                    sample.depends_on, sample_group, f"/entry/{sample.name}"
                )
                self._writer.add_dataset(sample_group, "depends_on", data=depends_on)
            if sample.distance is not None:
                distance_ds = self._writer.add_dataset(
                    sample_group, "distance", data=sample.distance
                )
                if sample.distance_units is not None:
                    self._writer.add_attribute(
                        distance_ds, "units", sample.distance_units
                    )

            if sample.ub_matrix is not None:
                self._writer.add_dataset(
                    sample_group, "ub_matrix", data=sample.ub_matrix
                )

            if sample.orientation_matrix is not None:
                self._writer.add_dataset(
                    sample_group, "orientation_matrix", data=sample.orientation_matrix
                )

    def _write_source(self, parent_group: Union[h5py.Group, Dict]):
        for source in self._source:
            source_group = self._create_nx_class(source.name, "NXsource", parent_group)
            if source.depends_on is not None:
                if isinstance(source.depends_on, str):
                    depends_on = source.depends_on
                else:
                    depends_on = self._add_transformations_to_file(
                        source.depends_on, source_group, f"/entry/{source.name}"
                    )
                self._writer.add_dataset(source_group, "depends_on", data=depends_on)
            if source.distance is not None:
                distance_ds = self._writer.add_dataset(
                    source_group, "distance", data=source.distance
                )
                if source.distance_units is not None:
                    self._writer.add_attribute(
                        distance_ds, "units", source.distance_units
                    )

    def _write_instrument(
        self, parent_group: Union[h5py.Group, Dict]
    ) -> Union[h5py.Group, Dict]:
        instrument_group = self._create_nx_class(
            "instrument", "NXinstrument", parent_group
        )
        self._writer.add_dataset(instrument_group, "name", self._instrument_name)
        return instrument_group

    def _write_detectors(self, parent_group: Union[h5py.Group, Dict], parent_path: str):
        for detector_index, detector in enumerate(self._detectors):
            detector_name = f"detector_{detector_index}"
            detector_group = self._add_detector_group_to_file(
                detector, parent_group, detector_name
            )
            if detector.data is not None:
                da = detector.data
                ds = self._writer.add_dataset(detector_group, "data", data=da.values)
                self._writer.add_attribute(ds, "units", str(da.unit))
                axes = [dim if dim in da.coords else '.' for dim in da.dims]
                self._writer.add_attribute(detector_group, "axes", axes)
                for key, coord in da.coords.items():
                    ds = self._writer.add_dataset(
                        detector_group, key, data=coord.values
                    )
                    self._writer.add_attribute(ds, "units", str(coord.unit))

            if detector.event_data is not None:
                self._add_event_data_group_to_file(
                    detector.event_data, detector_group, "events"
                )
            if detector.log is not None:
                self._add_log_group_to_file(detector.log, detector_group)
            if detector.depends_on is not None:
                depends_on = self._add_transformations_to_file(
                    detector.depends_on,
                    detector_group,
                    f"{parent_path}/{detector_name}",
                )
                self._writer.add_dataset(detector_group, "depends_on", data=depends_on)

    def _write_choppers(self, parent_group: Union[h5py.Group, Dict]):
        for chopper in self._choppers:
            chopper_group = self._create_nx_class(
                chopper.name, "NXdisk_chopper", parent_group
            )
            distance_ds = self._writer.add_dataset(
                chopper_group, "distance", data=chopper.distance
            )
            rotation_ds = self._writer.add_dataset(
                chopper_group, "rotation_speed", data=chopper.rotation_speed
            )
            if chopper.distance_units is not None:
                self._writer.add_attribute(distance_ds, "units", chopper.distance_units)
            if chopper.rotation_units is not None:
                self._writer.add_attribute(rotation_ds, "units", chopper.rotation_units)

    def _write_event_data(self, parent_group: Union[h5py.Group, Dict]):
        for event_data_index, event_data in enumerate(self._event_data):
            self._add_event_data_group_to_file(
                event_data, parent_group, f"events_{event_data_index}"
            )

    def _write_monitors(self, parent_group: Union[h5py.Group, Dict]):
        for monitor in self._monitors:
            self._add_monitor_group_to_file(monitor, parent_group)

    def _add_monitor_group_to_file(self, monitor: Monitor, parent_group: h5py.Group):
        monitor_group = self._create_nx_class(monitor.name, "NXmonitor", parent_group)
        data_group = self._writer.add_dataset(monitor_group, "data", monitor.data)
        self._writer.add_attribute(data_group, "units", '')
        self._writer.add_attribute(
            data_group, "axes", ",".join(name for name, _ in monitor.axes)
        )

        if monitor.events:
            self._write_event_data_to_group(monitor_group, monitor.events)

        for axis_name, axis_data in monitor.axes:
            # We write event data (if exists) first - if we've already written event
            # data the event index will already have been created so we skip writing
            # it here.
            if not monitor.events or not axis_name == "event_index":
                ds = self._writer.add_dataset(monitor_group, axis_name, axis_data)
                self._writer.add_attribute(ds, "units", '')
        if monitor.depends_on is not None:
            if isinstance(monitor.depends_on, str):
                depends_on = monitor.depends_on
            else:
                depends_on = self._add_transformations_to_file(
                    monitor.depends_on, monitor_group, f"/{monitor.name}"
                )
            self._writer.add_dataset(monitor_group, "depends_on", data=depends_on)

    def _write_datas(self, parent_group: Union[h5py.Group, Dict]):
        for data in self._datas:
            self._add_data_group_to_file(data, parent_group)

    def _add_data_group_to_file(self, data: Data, parent_group: h5py.Group):
        da = data.data
        group = self._create_nx_class(data.name, "NXdata", parent_group)
        self._writer.add_attribute(group, "axes", da.dims)
        self._writer.add_attribute(group, "signal", "signal1")
        signal = self._writer.add_dataset(group, "signal1", da.values)
        self._writer.add_attribute(signal, "units", str(da.unit))
        # Note: We are deliberately NOT adding AXISNAME_indices attributes for the
        # coords, since these were added late to the Nexus standard and therefore we
        # also need to support loading without the attributes. The attribute should be
        # set manually by the user if desired.
        for name, coord in da.coords.items():
            ds = self._writer.add_dataset(group, name, coord.values)
            self._writer.add_attribute(ds, "units", str(coord.unit))
        if data.attrs is not None:
            for k, v in data.attrs.items():
                self._writer.add_attribute(group, k, v)

    def _write_logs(self, parent_group: Union[h5py.Group, Dict]):
        for log in self._logs:
            self._add_log_group_to_file(log, parent_group)

    def _add_event_data_group_to_file(
        self, data: EventData, parent_group: h5py.Group, group_name: str
    ):
        event_group = self._create_nx_class(group_name, "NXevent_data", parent_group)
        self._write_event_data_to_group(event_group, data)

    def _write_event_data_to_group(self, event_group: h5py.Group, data: EventData):
        if data.event_id is not None:
            self._writer.add_dataset(event_group, "event_id", data=data.event_id)
        if data.event_time_offset is not None:
            event_time_offset_ds = self._writer.add_dataset(
                event_group, "event_time_offset", data=data.event_time_offset
            )
            self._writer.add_attribute(
                event_time_offset_ds, "units", data.event_time_offset_unit
            )
        if data.event_time_zero is not None:
            event_time_zero_ds = self._writer.add_dataset(
                event_group, "event_time_zero", data=data.event_time_zero
            )
            self._writer.add_attribute(
                event_time_zero_ds, "units", data.event_time_zero_unit
            )
            self._writer.add_attribute(
                event_time_zero_ds, "offset", data.event_time_zero_offset
            )
        if data.event_index is not None:
            self._writer.add_dataset(event_group, "event_index", data=data.event_index)

    def _add_transformations_to_file(
        self, transform: Transformation, parent_group: h5py.Group, parent_path: str
    ) -> str:
        transform_chain = [transform]
        while transform.depends_on is not None and not isinstance(
            transform.depends_on, str
        ):
            transform_chain.append(transform.depends_on)
            transform = transform.depends_on

        transforms_group_name = "transformations"
        transforms_group = self._create_nx_class(
            "transformations", "NXtransformations", parent_group
        )
        transform_chain.reverse()
        depends_on_str = (
            transform.depends_on if isinstance(transform.depends_on, str) else None
        )
        transform_group_path = f"{parent_path}/{transforms_group_name}"
        for transform_number, transform in enumerate(transform_chain):
            if transform.time is not None:
                depends_on_str = self._add_transformation_as_log(
                    transform,
                    transform_number,
                    transforms_group,
                    transform_group_path,
                    depends_on_str,
                )
            else:
                depends_on_str = self._add_transformation_as_dataset(
                    transform,
                    transform_number,
                    transforms_group,
                    transform_group_path,
                    depends_on_str,
                )
        return depends_on_str

    def _add_transformation_as_dataset(
        self,
        transform: Transformation,
        transform_number: int,
        transforms_group: h5py.Group,
        group_path: str,
        depends_on: Optional[str],
    ) -> str:
        transform_name = f"transform_{transform_number}"
        added_transform = self._writer.add_dataset(
            transforms_group, f"transform_{transform_number}", data=transform.value
        )
        self._add_transform_attributes(added_transform, depends_on, transform)
        if transform.value_units is not None:
            self._writer.add_attribute(added_transform, "units", transform.value_units)
        return f"{group_path}/{transform_name}"

    def _add_log_group_to_file(self, log: Log, parent_group: h5py.Group) -> h5py.Group:
        log_group = self._create_nx_class(log.name, "NXlog", parent_group)
        if log.value is not None:
            value_ds = self._writer.add_dataset(log_group, "value", log.value)
            if log.value_units is not None:
                self._writer.add_attribute(value_ds, "units", log.value_units)
        if log.time is not None:
            time_ds = self._writer.add_dataset(log_group, "time", data=log.time)
            if log.time_units is not None:
                self._writer.add_attribute(time_ds, "units", log.time_units)
            if log.start_time is not None:
                self._writer.add_attribute(time_ds, "start", log.start_time)
            if log.scaling_factor is not None:
                self._writer.add_attribute(
                    time_ds, "scaling_factor", log.scaling_factor
                )
        return log_group

    def _add_transformation_as_log(
        self,
        transform: Transformation,
        transform_number: int,
        transforms_group: h5py.Group,
        group_path: str,
        depends_on: Optional[str],
    ) -> str:
        transform_name = f"transform_{transform_number}"
        added_transform = self._add_log_group_to_file(
            Log(
                transform_name,
                transform.value,
                transform.time,
                transform.value_units,
                transform.time_units,
            ),
            transforms_group,
        )
        self._add_transform_attributes(added_transform, depends_on, transform)
        return f"{group_path}/{transform_name}"

    def _add_detector_group_to_file(
        self, detector: Detector, parent_group: h5py.Group, group_name: str
    ) -> h5py.Group:
        detector_group = self._create_nx_class(group_name, "NXdetector", parent_group)
        if detector.detector_numbers is not None:
            self._writer.add_dataset(
                detector_group, "detector_number", detector.detector_numbers
            )
        for dataset_name, array in (
            ("x_pixel_offset", detector.x_offsets),
            ("y_pixel_offset", detector.y_offsets),
            ("z_pixel_offset", detector.z_offsets),
        ):
            if array is not None:
                offsets_ds = self._writer.add_dataset(
                    detector_group, dataset_name, array
                )
                if detector.offsets_unit is not None:
                    self._writer.add_attribute(
                        offsets_ds, "units", detector.offsets_unit
                    )
        return detector_group

    def _add_transform_attributes(
        self,
        added_transform: Union[h5py.Group, h5py.Dataset],
        depends_on: Optional[str],
        transform: Transformation,
    ):
        self._writer.add_attribute(added_transform, "vector", transform.vector)
        self._writer.add_attribute(
            added_transform, "transformation_type", transform.transform_type.value
        )
        if transform.offset is not None:
            self._writer.add_attribute(added_transform, "offset", transform.offset)
        if transform.offset_unit is not None:
            self._writer.add_attribute(
                added_transform, "offset_units", transform.offset_unit
            )
        if depends_on is not None:
            self._writer.add_attribute(added_transform, "depends_on", depends_on)
        else:
            self._writer.add_attribute(
                added_transform, "depends_on", "."
            )  # means end of chain

    def _create_nx_class(
        self, group_name: str, nx_class_name: str, parent: h5root
    ) -> h5py.Group:
        nx_class = self._writer.add_group(parent, group_name)
        self._writer.add_attribute(nx_class, "NX_class", nx_class_name)
        return nx_class

    def _write_streams(self, root: Union[h5py.File, Dict]):
        if isinstance(self._writer, JsonWriter):
            for stream in self._streams:
                self._writer.add_stream(root, stream)



import argparse
import os

from nbconvert.exporters import NotebookExporter
from traitlets.config import Config

c = Config()
c.TagRemovePreprocessor.remove_cell_tags = ("solution",)
c.TagRemovePreprocessor.enabled = True
c.NotebookExporter.preprocessors = ["nbconvert.preprocessors.TagRemovePreprocessor"]

parser = argparse.ArgumentParser(
    description='Strip cells tagged with "solution" from notebooks.'
)
parser.add_argument(
    'names',
    metavar='input_path',
    type=str,
    nargs='+',
    help='Paths of notebooks to convert.',
)
parser.add_argument('--output-dir', dest='path', help='Output folder')

args = parser.parse_args()

for name in args.names:
    print(f"Stripping solution cells from {name}")
    output = NotebookExporter(config=c).from_filename(name)
    with open(os.path.join(args.path, os.path.basename(name)), "w") as f:
        f.write(output[0])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock
import pytest
import scipp as sc
from scipp.testing import assert_identical

from scippneutron.tof import fakes, unwrap

sl = pytest.importorskip('sciline')


@pytest.fixture
def ess_10s_14Hz() -> fakes.FakeSource:
    return fakes.FakeSource(
        frequency=sc.scalar(14.0, unit='Hz'), run_length=sc.scalar(10.0, unit='s')
    )


@pytest.fixture
def ess_10s_7Hz() -> fakes.FakeSource:
    return fakes.FakeSource(
        frequency=sc.scalar(7.0, unit='Hz'), run_length=sc.scalar(10.0, unit='s')
    )


@pytest.fixture
def ess_pulse() -> fakes.FakePulse:
    return fakes.FakePulse(
        time_min=sc.scalar(0.0, unit='ms'),
        time_max=sc.scalar(3.0, unit='ms'),
        wavelength_min=sc.scalar(0.1, unit='angstrom'),
        wavelength_max=sc.scalar(10.0, unit='angstrom'),
    )


def test_frame_period_is_pulse_period_if_not_pulse_skipping() -> None:
    pl = sl.Pipeline(unwrap.unwrap_providers())
    period = sc.scalar(123.0, unit='ms')
    pl[unwrap.PulsePeriod] = period
    assert_identical(pl.compute(unwrap.FramePeriod), period)


@pytest.mark.parametrize('stride', [1, 2, 3, 4])
def test_frame_period_is_multiple_pulse_period_if_pulse_skipping(stride) -> None:
    pl = sl.Pipeline(unwrap.unwrap_providers())
    period = sc.scalar(123.0, unit='ms')
    pl[unwrap.PulsePeriod] = period
    pl[unwrap.PulseStride] = stride
    assert_identical(pl.compute(unwrap.FramePeriod), stride * period)


@pytest.mark.parametrize('stride', [1, 2, 3, 4])
def test_pulse_offset(stride) -> None:
    pl = sl.Pipeline(unwrap.unwrap_providers(pulse_skipping=True))
    period = sc.scalar(123.0, unit='ms')
    pl[unwrap.PulsePeriod] = period
    pl[unwrap.PulseStride] = stride
    start = sc.datetime('2020-01-01T00:00:00.0', unit='ns')
    time_zero = start + sc.arange('pulse', 0 * period, 100 * period, period).to(
        unit='ns', dtype='int64'
    )
    pl[unwrap.TimeZero] = time_zero

    result = pl.compute(unwrap.PulseOffset)
    assert_identical(
        result, (sc.arange('pulse', 0, 100, dtype='int64') % stride) * period
    )


def test_offset_from_wrapped() -> None:
    pl = sl.Pipeline(unwrap.unwrap_providers())
    period = sc.scalar(123.0, unit='ms')
    pl[unwrap.PulsePeriod] = period
    pl[unwrap.FrameBounds] = unwrap.FrameBounds(
        sc.DataGroup(time=sc.array(dims=['bound'], values=[0.01, 0.02], unit='s'))
    )
    wrapped_offset = sc.linspace('event', 0.0, 123.0, num=1001, unit='ms')
    pl[unwrap.PulseWrappedTimeOffset] = unwrap.PulseWrappedTimeOffset(wrapped_offset)
    offset = pl.compute(unwrap.DeltaFromWrapped)
    # Times below 10 ms (we currently cut at lower bound) should be offset by period.
    da = sc.DataArray(offset, coords={'time': wrapped_offset})
    assert sc.all(da['time', : 10 * sc.Unit('ms')].data == period.to(unit='s'))
    assert sc.all(da['time', 10 * sc.Unit('ms') :].data == sc.scalar(0.0, unit='s'))


def test_offset_from_wrapped_has_no_special_handling_for_out_of_period_events() -> None:
    pl = sl.Pipeline(unwrap.unwrap_providers())
    period = sc.scalar(123.0, unit='ms')
    pl[unwrap.PulsePeriod] = period
    pl[unwrap.FrameBounds] = unwrap.FrameBounds(
        sc.DataGroup(time=sc.array(dims=['bound'], values=[0.01, 0.02], unit='s'))
    )
    wrapped_offset = sc.linspace('event', -10000.0, 10000.0, num=10001, unit='ms')
    pl[unwrap.PulseWrappedTimeOffset] = unwrap.PulseWrappedTimeOffset(wrapped_offset)
    offset = pl.compute(unwrap.DeltaFromWrapped)
    da = sc.DataArray(offset, coords={'time': wrapped_offset})
    # Negative times and times > 123 ms are technically invalid, but it does not affect
    # unwrapping, so they should be left as-is.
    assert sc.all(da['time', : 10 * sc.Unit('ms')].data == period.to(unit='s'))
    assert sc.all(da['time', 10 * sc.Unit('ms') :].data == sc.scalar(0.0, unit='s'))


def test_unwrap_with_no_choppers(ess_10s_14Hz, ess_pulse) -> None:
    # At this small distance the frames are not overlapping (with the given wavelength
    # range), despite not using any choppers.
    distance = sc.scalar(10.0, unit='m')
    beamline = fakes.FakeBeamline(
        source=ess_10s_14Hz,
        pulse=ess_pulse,
        choppers={},  # no choppers
        monitors={'monitor': distance},
        detectors={},
    )
    mon, ref = beamline.get_monitor('monitor')

    pl = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers()
    )
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )
    pl[unwrap.Choppers] = {}
    pl[unwrap.Ltotal] = distance
    unwrapped = pl.compute(unwrap.UnwrappedData)
    # No unwrap is happening, frame does not overlap next pulse.
    assert (mon.coords['event_time_zero'] == unwrapped.bins.coords['pulse_time']).all()

    origin = pl.compute(unwrap.TimeOfFlightOrigin)
    assert_identical(origin.time, sc.scalar(0.0015, unit='s'))
    assert_identical(origin.distance, sc.scalar(0.0, unit='m'))

    result = pl.compute(unwrap.TofData)
    assert_identical(
        result.hist(tof=1000).sum('pulse'), ref.hist(tof=1000).sum('pulse')
    )


def test_unwrap_with_frame_overlap_raises(ess_10s_14Hz, ess_pulse) -> None:
    distance = sc.scalar(46.0, unit='m')
    beamline = fakes.FakeBeamline(
        source=ess_10s_14Hz,
        pulse=ess_pulse,
        choppers={},  # no choppers
        monitors={'monitor': distance},
        detectors={},
    )
    mon, _ = beamline.get_monitor('monitor')

    pl = sl.Pipeline(unwrap.unwrap_providers())
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )
    pl[unwrap.Choppers] = {}
    pl[unwrap.Ltotal] = distance
    with pytest.raises(ValueError, match='Frames are overlapping'):
        pl.compute(unwrap.UnwrappedData)


def test_standard_unwrap(ess_10s_14Hz, ess_pulse) -> None:
    distance = sc.scalar(46.0, unit='m')
    beamline = fakes.FakeBeamline(
        source=ess_10s_14Hz,
        pulse=ess_pulse,
        choppers=fakes.psc_choppers,
        monitors={'monitor': distance},
        detectors={},
        time_of_flight_origin='psc1',
    )
    mon, ref = beamline.get_monitor('monitor')

    pl = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers()
    )
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )
    pl[unwrap.Choppers] = fakes.psc_choppers
    pl[unwrap.SourceChopperName] = 'psc1'
    pl[unwrap.Ltotal] = distance
    result = pl.compute(unwrap.TofData)
    assert_identical(
        result.hist(tof=1000).sum('pulse'), ref.hist(tof=1000).sum('pulse')
    )


def test_standard_unwrap_histogram_mode(ess_10s_14Hz, ess_pulse) -> None:
    distance = sc.scalar(46.0, unit='m')
    beamline = fakes.FakeBeamline(
        source=ess_10s_14Hz,
        pulse=ess_pulse,
        choppers=fakes.psc_choppers,
        monitors={'monitor': distance},
        detectors={},
        time_of_flight_origin='psc1',
    )
    mon, ref = beamline.get_monitor('monitor')
    mon = (
        mon.hist(
            event_time_offset=sc.linspace(
                'event_time_offset', 0.0, 1000.0 / 14, num=1001, unit='ms'
            ).to(unit='s')
        )
        .sum('pulse')
        .rename(event_time_offset='time_of_flight')
    )

    pl = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers()
    )
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )
    pl[unwrap.Choppers] = fakes.psc_choppers
    pl[unwrap.SourceChopperName] = 'psc1'
    pl[unwrap.Ltotal] = distance
    result = pl.compute(unwrap.TofData)
    assert_identical(result.sum(), ref.sum())


def test_pulse_skipping_unwrap(ess_10s_7Hz, ess_pulse) -> None:
    # Pretend pulse-skipping by using source of different frequency. In reality this
    # would be done using a chopper.
    distance = sc.scalar(46.0, unit='m')
    beamline = fakes.FakeBeamline(
        source=ess_10s_7Hz,
        pulse=ess_pulse,
        choppers=fakes.psc_choppers,
        monitors={'monitor': distance},
        detectors={},
        time_of_flight_origin='psc1',
    )
    mon, ref = beamline.get_monitor('monitor')

    pl = sl.Pipeline(
        unwrap.unwrap_providers(pulse_skipping=True)
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers()
    )
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = 0.5 * beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )

    pl[unwrap.Choppers] = fakes.psc_choppers
    pl[unwrap.SourceChopperName] = 'psc1'
    pl[unwrap.Ltotal] = distance
    pl[unwrap.PulseStride] = 2

    result = pl.compute(unwrap.TofData)
    assert_identical(result.sum(), ref.sum())
    assert_identical(
        result.hist(tof=1000).sum('pulse'), ref.hist(tof=1000).sum('pulse')
    )


def test_pulse_skipping_unwrap_histogram_mode_not_implemented(
    ess_10s_7Hz, ess_pulse
) -> None:
    distance = sc.scalar(46.0, unit='m')
    beamline = fakes.FakeBeamline(
        source=ess_10s_7Hz,
        pulse=ess_pulse,
        choppers=fakes.psc_choppers,
        monitors={'monitor': distance},
        detectors={},
        time_of_flight_origin='psc1',
    )
    mon, ref = beamline.get_monitor('monitor')
    mon = (
        mon.hist(
            event_time_offset=sc.linspace(
                'event_time_offset', 0.0, 1000.0 / 14, num=1001, unit='ms'
            ).to(unit='s')
        )
        .sum('pulse')
        .rename(event_time_offset='time_of_flight')
    )

    pl = sl.Pipeline(
        unwrap.unwrap_providers(pulse_skipping=True)
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers()
    )
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = 0.5 * beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )
    pl[unwrap.Choppers] = fakes.psc_choppers
    pl[unwrap.SourceChopperName] = 'psc1'
    pl[unwrap.Ltotal] = distance
    pl[unwrap.PulseStride] = 2
    with pytest.raises(NotImplementedError):
        pl.compute(unwrap.TofData)


def test_wfm_unwrap(ess_10s_14Hz, ess_pulse) -> None:
    distance = sc.scalar(20.0, unit='m')
    choppers = fakes.wfm_choppers.copy()
    # We currently have no way of detecting which cutouts of the "source" chopper
    # are used (not blocked by other choppers), for this test we remove those we
    # know are not used.
    choppers['wfm1'] = choppers['wfm1'][2:]
    beamline = fakes.FakeBeamline(
        source=ess_10s_14Hz,
        pulse=ess_pulse,
        choppers=choppers,
        monitors={'monitor': distance},
        detectors={},
        # time_of_flight_origin='wfm1',
    )
    mon, ref = beamline.get_monitor('monitor')

    pl = sl.Pipeline(
        unwrap.unwrap_providers()
        + unwrap.time_of_flight_providers()
        + unwrap.time_of_flight_origin_from_choppers_providers(wfm=True)
    )
    pl[unwrap.RawData] = mon
    pl[unwrap.PulsePeriod] = beamline._source.pulse_period
    pl[unwrap.SourceTimeRange] = ess_pulse.time_min, ess_pulse.time_max
    pl[unwrap.SourceWavelengthRange] = (
        ess_pulse.wavelength_min,
        ess_pulse.wavelength_max,
    )
    pl[unwrap.Choppers] = choppers
    # Actually the source should be at the center between wfm1 and wfm2, but we
    # currently don't have a way of handling this. We could manually create a
    # "virtual" chopper.
    pl[unwrap.SourceChopperName] = 'wfm1'
    pl[unwrap.Ltotal] = distance
    bounds = pl.compute(unwrap.SubframeBounds)
    assert bounds.sizes == {'bound': 2, 'subframe': 6}
    result = pl.compute(unwrap.TofData)
    ref.coords['Ltotal'] = distance - choppers['wfm1'].distance
    # FakeBeamline does not support WFM yet, we cannot run a better check for now.
    assert_identical(result.sum(), ref.sum())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
"""Utilities for logging in ScippNeutron.

The object returned by :func:`scippneutron.get_logger` is the only logger
used by ScippNeutron. ScippNeutron does not configure it in any way.
You are free to do so.
See also the `logging documentation <https://scipp.github.io/reference/logging.html>`_
for Scipp.
"""

import logging


def logger_name() -> str:
    """Return the name of ScippNeutron's logger."""
    return "scipp.neutron"


def get_logger() -> logging.Logger:
    """Return the logger used by ScippNeutron."""
    return logging.getLogger(logger_name())



import matplotlib
import numpy as np
import pytest
import scipp as sc

import scippneutron as scn

matplotlib.use('Agg')


def make_dataset_with_beamline():
    positions = [[1, 0, 0], [0, 1, 0], [0, 0, 1], [-1, 0, 0]]
    d = sc.Dataset(
        data={'a': sc.Variable(dims=['position', 'tof'], values=np.random.rand(4, 9))},
        coords={
            'tof': sc.Variable(
                dims=['tof'], values=np.arange(1000.0, 1010.0), unit=sc.units.us
            ),
            'position': sc.vectors(
                dims=['position'], values=positions, unit=sc.units.m
            ),
        },
    )

    d.coords['source_position'] = sc.vector(
        value=np.array([0, 0, -10]), unit=sc.units.m
    )
    d.coords['sample_position'] = sc.vector(value=np.array([0, 0, 0]), unit=sc.units.m)
    return d


def test_neutron_instrument_view_3d():
    d = make_dataset_with_beamline()
    scn.instrument_view(d["a"])


def test_neutron_instrument_view_with_masks():
    d = make_dataset_with_beamline()
    x = np.transpose(d.coords['position'].values)[0, :]
    d['a'].masks['amask'] = sc.Variable(
        dims=['position'], values=np.less(np.abs(x), 0.5)
    )
    scn.instrument_view(d["a"])


def test_neutron_instrument_view_with_cmap_args():
    d = make_dataset_with_beamline()
    scn.instrument_view(
        d["a"],
        vmin=0.001 * sc.units.one,
        vmax=5.0 * sc.units.one,
        cmap="magma",
        norm="log",
    )


def _make_component_settings(
    *,
    data,
    center='sample_position',
    type='box',
    size_unit=sc.units.m,
    wireframe=False,
    component_size=(0.1, 0.1, 0.1),
):
    comp_size = (
        sc.vector(value=component_size, unit=size_unit)
        if isinstance(component_size, tuple)
        else sc.scalar(value=component_size, unit=size_unit)
    )
    center = data.meta[center] if isinstance(center, str) else center
    sample_settings = {'center': center, 'size': comp_size, 'type': type}
    return sample_settings


def test_neutron_instrument_view_components_valid():
    d = make_dataset_with_beamline()
    scn.instrument_view(d["a"], components={'sample': _make_component_settings(data=d)})


def test_neutron_instrument_view_with_size_scalar():
    d = make_dataset_with_beamline()
    scn.instrument_view(
        d["a"],
        components={'sample': _make_component_settings(data=d, component_size=2.0)},
    )


def test_neutron_instrument_view_components_multiple_valid():
    d = make_dataset_with_beamline()
    scn.instrument_view(
        d["a"],
        components={
            'sample': _make_component_settings(data=d, center='sample_position'),
            'source': _make_component_settings(data=d, center='source_position'),
        },
    )


def test_neutron_instrument_view_components_with_non_beamline_component():
    d = make_dataset_with_beamline()
    widget_center = sc.vector(value=[1, 1, 1], unit=sc.units.m)
    scn.instrument_view(
        d["a"],
        components={'widget': _make_component_settings(data=d, center=widget_center)},
    )


def test_neutron_instrument_view_components_with_wireframe():
    d = make_dataset_with_beamline()
    scn.instrument_view(
        d["a"], components={'sample': _make_component_settings(data=d, wireframe=False)}
    )
    scn.instrument_view(
        d["a"], components={'sample': _make_component_settings(data=d, wireframe=True)}
    )


def test_neutron_instrument_view_components_with_invalid_type():
    d = make_dataset_with_beamline()
    # Check that all our valid shape types work
    for shape_type in ['box', 'cylinder', 'disk']:
        scn.instrument_view(
            d["a"],
            components={'sample': _make_component_settings(data=d, type=shape_type)},
        )
    with pytest.raises(ValueError):
        scn.instrument_view(
            d["a"],
            components={
                'sample': _make_component_settings(data=d, type='trefoil_knot')
            },
        )


def test_neutron_instrument_view_components_with_invalid_size_unit():
    d = make_dataset_with_beamline()
    # mm is fine, source_position is set in meters. Just a scaling factor.
    scn.instrument_view(
        d["a"],
        components={'sample': _make_component_settings(data=d, size_unit=sc.units.mm)},
    )
    # cannot scale us to meters. This should throw
    with pytest.raises(sc.UnitError):
        scn.instrument_view(
            d["a"],
            components={
                'sample': _make_component_settings(data=d, size_unit=sc.units.us)
            },
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)
# @author Jan-Lukas Wynen
"""Functions for computing coordinates in time-of-flight neutron scattering.

Some functions in this module use the neutron mass :math:`m_n`
and Planck constant :math:`h`.
Their values are taken from :mod:`scipp.constants`.
"""

from typing import Tuple

import numpy as np
import scipp as sc
import scipp.constants as const
from scipp.typing import Variable, VariableLike

from .._utils import as_float_type, elem_dtype, elem_unit


def _common_dtype(a, b):
    """Very limited type promotion.
    Only useful to check if the combination of a and b results in
    single or double precision float.
    """
    if elem_dtype(a) == sc.DType.float32 and elem_dtype(b) == sc.DType.float32:
        return sc.DType.float32
    return sc.DType.float64


def wavelength_from_tof(*, tof: Variable, Ltotal: Variable) -> Variable:
    r"""Compute the wavelength from time-of-flight.

    The result is the de Broglie wavelength

    .. math::

        \lambda = \frac{h t}{m_n L_\mathsf{total}}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    Ltotal:
        Total beam length.

    Returns
    -------
    :
        Wavelength :math:`\lambda`.
        Has unit ångström.
    """
    c = sc.to_unit(
        const.h / const.m_n,
        sc.units.angstrom * elem_unit(Ltotal) / elem_unit(tof),
        copy=False,
    )
    return as_float_type(c / Ltotal, tof) * tof


def dspacing_from_tof(
    *, tof: Variable, Ltotal: Variable, two_theta: Variable
) -> Variable:
    r"""Compute the d-spacing from time-of-flight.

    The result is the inter-planar lattice spacing

    .. math::

        d = \frac{h t}{m_n L_\mathsf{total}\; 2 \sin \theta}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    Ltotal:
        Total beam length.
    two_theta:
        Scattering angle :math:`2 \theta`.

    Returns
    -------
    :
        Inter-planar lattice spacing :math:`d`.
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definitions of ``two_theta`` and ``Ltotal``.
    """
    c = sc.to_unit(
        2 * const.m_n / const.h,
        elem_unit(tof) / sc.units.angstrom / elem_unit(Ltotal),
        copy=False,
    )
    return 1 / as_float_type(c * Ltotal * sc.sin(two_theta / 2), tof) * tof


def _energy_constant(energy_unit: sc.Unit, tof: Variable, length: Variable):
    return sc.to_unit(
        const.m_n / 2,
        energy_unit * (elem_unit(tof) / elem_unit(length)) ** 2,
        copy=False,
    )


def energy_from_tof(*, tof: Variable, Ltotal: Variable) -> Variable:
    r"""Compute the neutron energy from time-of-flight.

    The result is

    .. math::

        E = \frac{m_n L_\mathsf{total}^2}{2 t^2}

    Where :math:`m_n` is the neutron mass.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    Ltotal:
        Total beam length.
    Returns
    -------
    :
        Neutron energy :math:`E`.
        Has unit meV.
    """
    c = _energy_constant(sc.units.meV, tof, Ltotal)
    return as_float_type(c * Ltotal**2, tof) / tof ** sc.scalar(
        2, dtype=elem_dtype(tof)
    )


def _energy_transfer_t0(energy, tof, length):
    dtype = _common_dtype(energy, tof)
    c = as_float_type(_energy_constant(elem_unit(energy), tof, length), energy)
    return length.astype(dtype, copy=False) * sc.sqrt(c / energy)


def energy_transfer_direct_from_tof(
    *, tof: Variable, L1: Variable, L2: Variable, incident_energy: Variable
) -> Variable:
    r"""Compute the energy transfer in direct inelastic scattering.

    The result is

    .. math::

        \Delta E = E_i - \frac{m_n L_2^2}{2 {(t - t_0)}^2}

    With

    .. math::

        t_0 = \sqrt{m_n L_1^2 / (2 E_i)}

    and :math:`m_n` the neutron mass.

    The result is ``NaN`` for unphysical points, that is, where :math:`t < t_0`.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    L1:
        Primary beam length.
    L2:
        Secondary beam length.
    incident_energy:
        Energy before scattering :math:`E_i`.

    Returns
    -------
    :
        Energy transfer :math:`\Delta E`.
        Has the same unit as incident_energy.

    See Also
    --------
    scippneutron.conversions.tof.energy_transfer_indirect_from_tof
    """
    t0 = _energy_transfer_t0(incident_energy, tof, L1)
    c = _energy_constant(elem_unit(incident_energy), tof, L2)
    dtype = _common_dtype(incident_energy, tof)
    scale = (c * L2**2).astype(dtype, copy=False)
    delta_tof = tof - t0
    return sc.where(
        delta_tof <= sc.scalar(0, unit=elem_unit(delta_tof)),
        sc.scalar(np.nan, dtype=dtype, unit=elem_unit(incident_energy)),
        incident_energy - scale / delta_tof**2,
    )


def energy_transfer_indirect_from_tof(
    *, tof: Variable, L1: Variable, L2: Variable, final_energy: Variable
) -> Variable:
    r"""Compute the energy transfer in indirect inelastic scattering.

    The result is

    .. math::

        \Delta E = \frac{m_n L_1^2}{2 {(t - t_0)}^2} - E_f

    With

    .. math::

        t_0 = \sqrt{m_n L_2^2 / (2 E_f)}

    and :math:`m_n` the neutron mass.

    The result is ``NaN`` for unphysical points, that is, where :math:`t < t_0`.

    Parameters
    ----------
    tof:
        Time-of-flight :math:`t`.
    L1:
        Primary beam length.
    L2:
        Secondary beam length.
    final_energy:
        Energy after scattering :math:`E_f`.

    Returns
    -------
    :
        Energy transfer :math:`\Delta E`.
        Has the same unit as final_energy.

    See Also
    --------
    scippneutron.conversions.tof.energy_transfer_direct_from_tof
    """
    t0 = _energy_transfer_t0(final_energy, tof, L2)
    c = _energy_constant(elem_unit(final_energy), tof, L1)
    dtype = _common_dtype(final_energy, tof)
    scale = (c * L1**2).astype(dtype, copy=False)
    delta_tof = -t0 + tof  # Order chosen such that output.dims = ['spectrum', 'tof']
    return sc.where(
        delta_tof <= sc.scalar(0, unit=elem_unit(delta_tof)),
        sc.scalar(np.nan, dtype=dtype, unit=elem_unit(final_energy)),
        scale / delta_tof**2 - final_energy,
    )


def energy_from_wavelength(*, wavelength: Variable) -> Variable:
    r"""Compute the neutron energy from wavelength.

    The result is

    .. math::

        E = \frac{h^2}{2 m_n \lambda^2}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\lambda`.
        Has unit meV.

    Returns
    -------
    :
        Neutron energy :math:`E`.
    """
    c = as_float_type(
        sc.to_unit(
            const.h**2 / 2 / const.m_n, sc.units.meV * elem_unit(wavelength) ** 2
        ),
        wavelength,
    )
    return c / wavelength**2


def wavelength_from_energy(*, energy: Variable) -> Variable:
    r"""Compute the wavelength from the neutron energy.

    The result is the de Broglie wavelength

    .. math::

        \lambda = \frac{h}{\sqrt{2 m_n E}}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    energy:
        Neutron energy :math:`E`.

    Returns
    -------
    :
        Wavelength :math:`\lambda`.
        Has unit ångström.
    """
    c = as_float_type(
        sc.to_unit(
            const.h**2 / 2 / const.m_n, sc.units.angstrom**2 * elem_unit(energy)
        ),
        energy,
    )
    return sc.sqrt(c / energy)


def _wavelength_Q_conversions(x: Variable, two_theta: Variable) -> Variable:
    """Convert either from Q to wavelength or vice-versa."""
    c = as_float_type(4 * const.pi, x)
    return c * sc.sin(as_float_type(two_theta, x) / 2) / x


def Q_from_wavelength(*, wavelength: Variable, two_theta: Variable) -> Variable:
    r"""Compute the absolute value of the momentum transfer from wavelength.

    The result is

    .. math::

        Q = \frac{4 \pi \sin \theta}{\lambda}

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\lambda`.
    two_theta:
        Scattering angle :math:`2 \theta`.

    Returns
    -------
    :
        Momentum transfer :math:`Q`.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    return _wavelength_Q_conversions(wavelength, two_theta)


def wavelength_from_Q(*, Q: Variable, two_theta: Variable) -> Variable:
    r"""Compute the wavelength from momentum transfer.

    The result is the de Broglie wavelength

    .. math::

        \lambda = \frac{4 \pi \sin \theta}{Q}

    Parameters
    ----------
    Q:
        Momentum transfer.
    two_theta:
        Scattering angle :math:`2 \theta`.

    Returns
    -------
    :
        Wavelength :math:`\lambda`
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    return sc.to_unit(
        _wavelength_Q_conversions(Q, two_theta), unit='angstrom', copy=False
    )


def Q_elements_from_wavelength(
    *, wavelength: Variable, incident_beam: Variable, scattered_beam: Variable
) -> Tuple[Variable, Variable, Variable]:
    r"""Compute them momentum transfer vector from wavelength.

    Computes the three components of the Q-vector :math:`Q_x, Q_y, Q_z`
    separately using

    .. math::

        \vec{Q} &= (Q_x, Q_y, Q_z) \\
        \vec{Q} &= \vec{k}_i - \vec{k}_f
                 = \frac{2\pi}{\lambda} \left(\hat{e}_i - \hat{e}_f\right),

    where the unit vectors for incident momentum and final momentum

    .. math::

        \hat{e}_i &= \vec{k_i} / | \vec{k_i} | \\
        \hat{e}_f &= \vec{k_f} / | \vec{k_f} |

    are defined as the directions of ``incident_beam`` and ``scattered_beam``,
    respectively.

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\lambda`.
    incident_beam:
        Beam from source to sample. Expects ``dtype=vector3``.
    scattered_beam:
        Beam from sample to detector. Expects ``dtype=vector3``.

    Returns
    -------
    Qx: scipp.Variable
        x-component of the momentum transfer :math:`\vec{Q}`.
    Qy: scipp.Variable
        y-component of the momentum transfer :math:`\vec{Q}`.
    Qz: scipp.Variable
        z-component of the momentum transfer :math:`\vec{Q}`.
    """
    e_i = incident_beam / sc.norm(incident_beam)
    e_f = scattered_beam / sc.norm(scattered_beam)
    e = e_i - e_f
    k = 2 * np.pi / wavelength
    return k * e.fields.x, k * e.fields.y, k * e.fields.z


def dspacing_from_wavelength(*, wavelength: Variable, two_theta: Variable) -> Variable:
    r"""Compute the d-spacing from wavelength.

    The result is the inter-planar lattice spacing

    .. math::

        d = \frac{\lambda}{2 \sin \theta}

    Parameters
    ----------
    wavelength:
        De Broglie wavelength :math:`\lambda`.
    two_theta:
        Scattering angle :math:`2 \theta`.

    Returns
    -------
    :
        Inter-planar lattice spacing :math:`d`.
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    c = as_float_type(
        sc.scalar(0.5).to(unit=sc.units.angstrom / elem_unit(wavelength)), wavelength
    )
    return c * wavelength / sc.sin(as_float_type(two_theta, wavelength) / 2)


def dspacing_from_energy(*, energy: Variable, two_theta: Variable) -> Variable:
    r"""Compute the d-spacing from the neutron energy.

    The result is the inter-planar lattice spacing

    .. math::

        d = \frac{h}{\sqrt{8 m_n E} \sin \theta}

    Where :math:`m_n` is the neutron mass and :math:`h` the Planck constant.

    Parameters
    ----------
    energy:
        Neutron energy :math:`E`.
    two_theta:
        Scattering angle :math:`2 \theta`.

    Returns
    -------
    :
        Inter-planar lattice spacing :math:`d`.
        Has unit ångström.

    See Also
    --------
    scippneutron.conversions.beamline:
        Definition of ``two_theta``.
    """
    c = as_float_type(
        sc.to_unit(
            const.h**2 / 8 / const.m_n, sc.units.angstrom**2 * elem_unit(energy)
        ),
        energy,
    )
    return sc.sqrt(c / energy) / sc.sin(as_float_type(two_theta, energy) / 2)


def Q_vec_from_Q_elements(*, Qx: Variable, Qy: Variable, Qz: Variable) -> Variable:
    """Combine elements of Q into a single vector variable.

    Parameters
    ----------
    Qx:
        x-elements of the momentum transfer.
    Qy:
        y-elements of the momentum transfer.
    Qz:
        z-elements of the momentum transfer.

    Returns
    -------
    :
        ``Qx``, ``Qy``, ``Qz`` combined into a single variable of dtype ``vector3``.
    """
    if Qx.sizes != Qy.sizes or Qx.sizes != Qz.sizes:
        raise sc.DimensionError(
            "Qx, Qy, Qz must have the same sizes. "
            f"Got {Qx.sizes=}, {Qy.sizes=}, {Qz.sizes=}."
        )
    return sc.spatial.as_vectors(Qx, Qy, Qz)


def ub_matrix_from_u_and_b(*, u_matrix: Variable, b_matrix: Variable) -> Variable:
    r"""Compute the UB matrix from U and B matrices.

    .. math::

        \mathsf{UB} = U \cdot B

    where :math:`U` and :math:`B` are defined as in
    :cite:`busing:1967,mantid-lattice:2023`.

    Parameters
    ----------
    u_matrix:
        :math:`U`.
    b_matrix:
        :math:`B`.

    Returns
    -------
    :
        :math:`\mathsf{UB}`.
    """
    return u_matrix * b_matrix


def hkl_vec_from_Q_vec(
    *, Q_vec: Variable, ub_matrix: Variable, sample_rotation: Variable
) -> Variable:
    r"""Compute hkl indices from momentum transfer.

    The hkl indices define the components of the momentum transfer in the
    sample coordinate system

    .. math::

        \vec{Q} = \begin{pmatrix} h \\ k \\ l \end{pmatrix}.

    In the lab frame, the momentum transfer as computed by
    :func:`scippneutron.conversion.tof.Q_elements_from_wavelength`
    is defined as

    .. math::

        \vec{Q}_l = \vec{k}_i - \vec{k}_f .

    This quantity is called :math:`Q` elsewhere in ScippNeutron.

    Those two :math:`Q`'s are related by via

    .. math::

        \vec{Q}_l = 2 \pi R U B \vec{Q},

    where :math:`U` and :math:`B` transform from sample space to the lab frame.
    :math:`R` encodes the sample rotation, e.g., as given by a goniometer.
    See, e.g., Refs. :cite:`busing:1967,mantid-lattice:2023,savici:2011`
    for a definition.

    This function computes the elements of :math:`\vec{Q}`, :math:`h, k, l` by inverting
    the above equation.

    Parameters
    ----------
    Q_vec:
        Momentum transfer :math:`\vec{Q}_l` as a vector variable.
    ub_matrix:
        Matrix :math:`\mathsf{UB}`.
    sample_rotation:
        Sample rotation matrix :math:`R`.

    Returns
    -------
    :
        :math:`h, k, l` as a vector variable.

    See also
    --------
    scippneutron.conversion.tof.Q_elements_from_wavelength:
        Computes ``Q_l``.
    scippneutron.conversion.tof.Q_vec_from_Q_elements:
        Packs elements ``Qx``, ``Qy``, ``Qz`` into a single vector.
    scippneutron.conversion.tof.ub_matrix_from_u_and_b:
        Compute :math:`\mathsf{UB}` from :math:`B` and :math:`B` matrices.
    scippneutron.conversion.tof.hkl_elements_from_hkl_vec:
        Unpack the returned hkl vector.
    """
    # There are different ways to implement this with different performance and
    # accuracy characteristics.
    # Matrix-matrix products typically have worse accuracy than matrix-vector
    # products due to repeated floating point cutoffs.
    # Matrix inversions are often unstable; however, all matrices used here are small.
    # Potential implementations are
    #
    # (sc.spatial.inv(B) * (sc.spatial.inv(U) * (sc.spatial.inv(R) * Q))) / (2*np.pi)
    #
    # (((sc.spatial.inv(B) * sc.spatial.inv(U)) * sc.spatial.inv(R)) * Q) / (2*np.pi)
    #
    # (sc.spatial.inv((R * U) * B) * Q) / (2*np.pi)
    #
    # All 3 were tested with random matrices and vectors and the first was found
    # to have the best overall accuracy, being an order of magnitude better than the
    # others in particularly bad cases and equal in the majority of cases.
    #
    # Concerning performance, R, U, B are scalar variables or short array variables
    # in typical use cases while Q is typically a long, potentially multi-dim array.
    # So implementation 3 will likely perform the best.
    #
    # This function uses implementation 3 as the performance gain
    # is expected to be significant over 1 and 2.
    return (sc.spatial.inv(sample_rotation * ub_matrix) * Q_vec) / (2 * np.pi)


def hkl_elements_from_hkl_vec(
    *, hkl_vec: Variable
) -> Tuple[Variable, Variable, Variable]:
    """Unpack vector of hkl indices into separate variables.

    Parameters
    ----------
    hkl_vec:
        Vector of hkl indices.

    Returns
    -------
    h: scipp.Variable
        1st component of the hkl vector.
    k: scipp.Variable
        2nd component of the hkl vector.
    l: scipp.Variable
        3rd component of the hkl vector.
    """
    return hkl_vec.fields.x, hkl_vec.fields.y, hkl_vec.fields.z


def time_at_sample_from_tof(
    *,
    pulse_time: VariableLike,
    tof: VariableLike,
    L2: VariableLike,
    wavelength: VariableLike,
) -> VariableLike:
    """Compute the absolute time when the neutron passed through the sample.

    The result is

    .. math::

        t_{sample} = t_{pulse} + t_{of} - L_2 / v

    where

    .. math::

        v = \\frac{h}{m_n \\lambda}

    where :math:`v` is the estimated velocity of the neutron over its path from
    sample to detector and :math:`\\lambda` is the wavelength of the neutron.

    Parameters
    ----------
    pulse_time:
        absolute time when time of flight is 0
    tof:
        the time of fligth of the neutron
    L2:
        path length from sample to detector
    wavelength:
        wavelength of the neutron (at the detector).
        Assuming this did not change during the neutrons
        travel from sample to detector this can be used to compute the
        velocity the neutron had between the sample and the detector.

    Returns
    -------
    :
        :math:`t_{sample}`
    """
    c = sc.to_unit(
        const.h / const.m_n,
        sc.units.angstrom * elem_unit(L2) / elem_unit(tof),
        copy=False,
    )
    return pulse_time + tof - L2 * wavelength / c
